"""
Generate Gauss-Kronrod quadrature abscissas and weights.

Example usage
-------------

Generate Gauss-Kronrod quadrature rules::

    >>> distribution = chaospy.Beta(2, 2, lower=-1, upper=1)
    >>> for order in range(5):  # doctest: +NORMALIZE_WHITESPACE
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="kronrod")
    ...     print(abscissas.round(2), weights.round(2))
    [[0.]] [1.]
    [[-0.65 -0.    0.65]] [0.23 0.53 0.23]
    [[-0.82 -0.45  0.    0.45  0.82]] [0.07 0.26 0.34 0.26 0.07]
    [[-0.89 -0.65 -0.34 -0.    0.34  0.65  0.89]]
     [0.03 0.12 0.22 0.26 0.22 0.12 0.03]
    [[-0.93 -0.77 -0.54 -0.29 -0.    0.29  0.54  0.77  0.93]]
     [0.01 0.06 0.13 0.19 0.22 0.19 0.13 0.06 0.01]

Compare Gauss-Kronrod builds on top of Gauss-Legendre quadrature to pure
Gauss-Legendre::

    >>> distribution = chaospy.Uniform(-1, 1)
    >>> for order in range(5):
    ...     abscissas1, weights1 = chaospy.generate_quadrature(
    ...         order, distribution, rule="gaussian")
    ...     abscissas2, weights2 = chaospy.generate_quadrature(
    ...         order+1, distribution, rule="kronrod")
    ...     print(abscissas1.round(2), abscissas2[:, 1::2].round(2))
    [[0.]] [[-0.]]
    [[-0.58  0.58]] [[-0.58  0.58]]
    [[-0.77 -0.    0.77]] [[-0.77 -0.    0.77]]
    [[-0.86 -0.34  0.34  0.86]] [[-0.86 -0.34  0.34  0.86]]
    [[-0.91 -0.54  0.    0.54  0.91]] [[-0.91 -0.54 -0.    0.54  0.91]]

Gauss-Kronrod build on top of Gauss-Hermite quadrature::

    >>> distribution = chaospy.Normal(0, 1)
    >>> for order in range(2):
    ...     abscissas1, weights1 = chaospy.generate_quadrature(
    ...         order, distribution, rule="gaussian")
    ...     abscissas2, weights2 = chaospy.generate_quadrature(
    ...         order+1, distribution, rule="kronrod")
    ...     print(abscissas1.round(2), abscissas2.round(2))
    [[0.]] [[-1.73  0.    1.73]]
    [[-1.  1.]] [[-2.45 -1.    0.    1.    2.45]]

Applying Gauss-Kronrod to Gauss-Hermite quadrature, for an order known to not
exist::

    >>> chaospy.generate_quadrature(  # doctest: +IGNORE_EXCEPTION_DETAIL
    ...     5, distribution, rule="kronrod")
    Traceback (most recent call last):
        ...
    numpy.linalg.LinAlgError: \
Invalid recurrence coefficients can not be used for constructing Gaussian quadrature rule

Multivariate support::

    >>> distribution = chaospy.J(
    ...     chaospy.Uniform(0, 1), chaospy.Beta(4, 5))
    >>> abscissas, weights = chaospy.generate_quadrature(
    ...     2, distribution, rule="kronrod")
    >>> abscissas.round(3)
    array([[0.037, 0.037, 0.037, 0.037, 0.037, 0.211, 0.211, 0.211, 0.211,
            0.211, 0.5  , 0.5  , 0.5  , 0.5  , 0.5  , 0.789, 0.789, 0.789,
            0.789, 0.789, 0.963, 0.963, 0.963, 0.963, 0.963],
           [0.144, 0.297, 0.444, 0.612, 0.796, 0.144, 0.297, 0.444, 0.612,
            0.796, 0.144, 0.297, 0.444, 0.612, 0.796, 0.144, 0.297, 0.444,
            0.612, 0.796, 0.144, 0.297, 0.444, 0.612, 0.796]])
    >>> weights.round(3)
    array([0.006, 0.027, 0.035, 0.026, 0.004, 0.016, 0.067, 0.086, 0.065,
           0.011, 0.02 , 0.085, 0.11 , 0.083, 0.014, 0.016, 0.067, 0.086,
           0.065, 0.011, 0.006, 0.027, 0.035, 0.026, 0.004])
"""
from __future__ import division
import math

import numpy
import chaospy

from .utils import combine_quadrature


def kronrod(
        order,
        dist,
        recurrence_algorithm="stieltjes",
        rule="clenshaw_curtis",
        tolerance=1e-10,
        scaling=3,
        n_max=5000,
):
    """
    Generate Gauss-Kronrod quadrature abscissas and weights.

    Gauss-Kronrod quadrature is an adaptive method for Gaussian quadrature
    rule. It builds on top of other quadrature rules by extending "pure"
    Gaussian quadrature rules with extra abscissas and new weights such that
    already used abscissas can be reused. For more details, see `Wikipedia
    article
    <https://en.wikipedia.org/wiki/Gauss%E2%80%93Kronrod_quadrature_formula>`_.

    For each order ``N`` taken with ordinary Gaussian quadrature, Gauss-Kronrod
    will create ``2N+1`` abscissas where all of the ``N`` "old" abscissas are
    all interlaced between the "new" ones.

    The algorithm is well suited for any Jacobi scheme, i.e. quadrature
    involving Uniform or Beta distribution, and might work on others as well.
    However, it will not work everywhere. For example `Kahaner and Monegato
    <https://link.springer.com/article/10.1007/BF01590820>`_ showed that higher
    order Gauss-Kronrod quadrature for Gauss-Hermite and Gauss-Laguerre does
    not exist.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Distribution):
            The distribution which density will be used as weight function.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        tolerance (float):
            The allowed relative error in norm between two quadrature orders
            before method assumes convergence.
        scaling (float):
            A multiplier the adaptive order increases with for each step
            quadrature order is not converged. Use 0 to indicate unit
            increments.
        n_max (int):
            The allowed number of quadrature points to use in approximation.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Raises:
        ValueError:
            Error raised if Kronrod algorithm results in negative recurrence
            coefficients.

    Notes:
        Code is adapted from `quadpy <https://github.com/nschloe/quadpy>`_,
        which adapted his code from `W. Gautschi
        <https://www.cs.purdue.edu/archives/2002/wxg/codes/OPQ.html>`_.
        Algorithm for calculating Kronrod-Jacobi matrices was first published
        Algorithm as proposed by Laurie :cite:`laurie_calculation_1997`.

    Example:
        >>> distribution = chaospy.Uniform(-1, 1)
        >>> abscissas, weights = chaospy.quadrature.kronrod(4, distribution)
        >>> abscissas.round(2)
        array([[-0.98, -0.86, -0.64, -0.34,  0.  ,  0.34,  0.64,  0.86,  0.98]])
        >>> weights.round(3)
        array([0.031, 0.085, 0.133, 0.163, 0.173, 0.163, 0.133, 0.085, 0.031])

    """
    length = int(numpy.ceil(3*(order+1) / 2.0))
    coefficients = chaospy.construct_recurrence_coefficients(
        order=length,
        dist=dist,
        recurrence_algorithm=recurrence_algorithm,
        rule=rule,
        tolerance=tolerance,
        scaling=scaling,
        n_max=n_max,
    )

    coefficients = [kronrod_jacobi(order, coeffs) for coeffs in coefficients]

    abscissas, weights = chaospy.coefficients_to_quadrature(coefficients)
    return combine_quadrature(abscissas, weights)


def kronrod_jacobi(order, coeffs):
    """
    Create the three-terms-recursion coefficients resulting from the
    Kronrod-Jacobi matrix.

    Augment three terms recurrence coefficients to add extra Gauss-Kronrod
    terms.

    Args:
        order (int):
            Order of the Gaussian quadrature rule.
        coeffs (numpy.ndarray):
            Three terms recurrence coefficients of the Gaussian quadrature
            rule.

    Returns:
        Three terms recurrence coefficients of the Gauss-Kronrod quadrature
        rule.
    """
    if not order:
        return kronrod_jacobi(1, coeffs)[:, :1]

    bound = int(math.floor(3*order/2.0))+1
    coeffs_a = numpy.zeros(2*order+1)
    coeffs_a[:bound] = coeffs[0][:bound]

    bound = int(math.ceil(3*order/2.0))+1
    coeffs_b = numpy.zeros(2*order+1)
    coeffs_b[:bound] = coeffs[1][:bound]

    sigma = numpy.zeros((2, order//2+2))
    sigma[1, 1] = coeffs_b[order+1]

    for idx in range(order-1):
        idy = numpy.arange((idx+1)//2, -1, -1)
        sigma[0, idy+1] = numpy.cumsum(
            (coeffs_a[idy+order+1]-coeffs_a[idx-idy])*sigma[1, idy+1]+
            coeffs_b[idy+order+1]*sigma[0, idy]-
            coeffs_b[idx-idy]*sigma[0, idy+1]
        )
        sigma = numpy.roll(sigma, 1, axis=0)

    sigma[0, 1:order//2+2] = sigma[0, :order//2+1]
    for idx in range(order-1, 2*order-2):
        idy = numpy.arange(idx-order+1, (idx-1)//2+1)
        j = order-1-idx+idy
        sigma[0, j+1] = numpy.cumsum(
            -(coeffs_a[idy+order+1]-coeffs_a[idx-idy])*sigma[1, j+1]-
            coeffs_b[idy+order+1]*sigma[0, j+1]+
            coeffs_b[idx-idy]*sigma[0, j+2]
        )
        j = j[-1]
        idy = (idx+1)//2
        if idx%2 == 0:
            coeffs_a[idy+order+1] = (
                coeffs_a[idy]+
                (sigma[0, j+1]-coeffs_b[idy+order+1]*sigma[0, j+2])/
                sigma[1, j+2]
            )
        else:
            coeffs_b[idy+order+1] = sigma[0, j+1]/sigma[0, j+2]
        sigma = numpy.roll(sigma, 1, axis=0)

    coeffs_a[2*order] = (coeffs_a[order-1]-
                         coeffs_b[2*order]*sigma[0, 1]/sigma[1, 1])
    return numpy.asfarray([coeffs_a, coeffs_b])



"""Collection of three terms recurrence algorithms."""
from .frontend import (
    construct_recurrence_coefficients, RECURRENCE_ALGORITHMS)
from .jacobi import coefficients_to_quadrature

from .chebyshev import modified_chebyshev
from .lanczos import lanczos
from .stieltjes import stieltjes, discretized_stieltjes, analytical_stieltjes



"""Auto correlation function."""
import numpy

from .pearson import Corr


def Acf(poly, dist, n_steps=None, **kws):
    """
    Auto-correlation function.

    Args:
        poly (numpoly.ndpoly):
            Polynomial of interest. Must have ``len(poly) > n_steps``.
        dist (Distribution):
            Defines the space the correlation is taken on.
        n_steps (int):
            The number of time steps apart included. If omitted set to
            ``len(poly)/2+1``.

    Returns:
        (numpy.ndarray) :
            Auto-correlation of ``poly`` with shape ``(n_steps,)``. Note that
            by definition ``Q[0]=1``.

    Examples:
        >>> poly = chaospy.monomial(1, 10)
        >>> dist = chaospy.Uniform()
        >>> chaospy.Acf(poly, dist).round(4)
        array([1.    , 0.9915, 0.9722, 0.9457, 0.9127])

    """
    n_steps = int(len(poly)/2+1) if n_steps is None else n_steps
    correlation = Corr(poly, dist, **kws)
    return numpy.array([numpy.mean(correlation.diagonal(idx), axis=0)
                        for idx in range(n_steps)])



"""Collection of low-discrepancy sequences."""
from .additive_recursion import create_additive_recursion_samples
from .chebyshev import create_chebyshev_samples, create_nested_chebyshev_samples
from .grid import create_grid_samples, create_nested_grid_samples
from .halton import create_halton_samples
from .hammersley import create_hammersley_samples
from .sobol import create_sobol_samples
from .korobov import create_korobov_samples



"""
Generates samples from the Sobol sequence.

Papers::

    Antonov, Saleev,
    USSR Computational Mathematics and Mathematical Physics,
    Volume 19, 1980, pages 252 - 256.

    Paul Bratley, Bennett Fox,
    Algorithm 659:
    Implementing Sobol's Quasirandom Sequence Generator,
    ACM Transactions on Mathematical Software,
    Volume 14, Number 1, pages 88-100, 1988.

    Bennett Fox,
    Algorithm 647:
    Implementation and Relative Efficiency of Quasirandom
    Sequence Generators,
    ACM Transactions on Mathematical Software,
    Volume 12, Number 4, pages 362-376, 1986.

    Ilya Sobol,
    USSR Computational Mathematics and Mathematical Physics,
    Volume 16, pages 236-242, 1977.

    Ilya Sobol, Levitan,
    The Production of Points Uniformly Distributed in a Multidimensional
    Cube (in Russian),
    Preprint IPM Akad. Nauk SSSR,
    Number 40, Moscow 1976.

"""
import math

import numpy

from .sobol_constants import DIM_MAX, LOG_MAX, POLY, SOURCE_SAMPLES


def create_sobol_samples(order, dim, seed=1):
    """
    Generates samples from the Sobol sequence.

    Sobol sequences (also called LP_T sequences or (t, s) sequences in base 2)
    are an example of quasi-random low-discrepancy sequences. They were first
    introduced by the Russian mathematician Ilya M. Sobol in 1967.

    These sequences use a base of two to form successively finer uniform
    partitions of the unit interval and then reorder the coordinates in each
    dimension.

    Args:
        order (int):
            Number of unique samples to generate.
        dim (int):
            Number of spacial dimensions. Must satisfy ``0 < dim < 1111``.
        seed (int):
            Starting seed. Non-positive values are treated as 1. If omitted,
            consecutive samples are used.

    Returns:
        (numpy.ndarray):
            Quasi-random vector with ``shape == (dim, order)``.

    Notes:
        Implementation based on the initial work of Sobol
        :cite:`sobol_distribution_1967`. This implementation is based on the
        work of `Burkardt
        <https://people.sc.fsu.edu/~jburkardt/py_src/sobol/sobol.html>`_.

    Examples:
        >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
        >>> samples = distribution.sample(3, rule="sobol")
        >>> samples.round(4)
        array([[0.5 , 0.75, 0.25],
               [0.5 , 0.25, 0.75]])
        >>> samples = distribution.sample(4, rule="sobol")
        >>> samples.round(4)
        array([[0.5  , 0.75 , 0.25 , 0.375],
               [0.5  , 0.25 , 0.75 , 0.375]])

    """
    assert 0 < dim < DIM_MAX, "dim in [1, 1111]"

    # Initialize row 1 of V.
    samples = SOURCE_SAMPLES.copy()
    samples[0, 0:LOG_MAX] = 1

    # Initialize the remaining rows of V.
    for idx in range(1, dim):

        # The bits of the integer POLY(I) gives the form of polynomial:
        degree = int(math.log(POLY[idx], 2))

        # Expand this bit pattern to separate components:
        includ = numpy.array([val == "1" for val in bin(POLY[idx])[-degree:]])

        #Calculate the remaining elements of row I as explained
        #in Bratley and Fox, section 2.
        for idy in range(degree+1, LOG_MAX+1):
            newv = samples[idx, idy-degree-1].item()
            base = 1
            for idz in range(1, degree+1):
                base *= 2
                if includ[idz-1]:
                    newv = newv ^ base * samples[idx, idy-idz-1].item()
            samples[idx, idy-1] = newv

    samples = samples[:dim]

    # Multiply columns of V by appropriate power of 2.
    samples *= 2**(numpy.arange(LOG_MAX, 0, -1, dtype=int))

    #RECIPD is 1/(common denominator of the elements in V).
    recipd = 0.5**(LOG_MAX+1)
    lastq = numpy.zeros(dim, dtype=int)

    seed = int(seed) if seed > 1 else 1

    for seed_ in range(seed):
        lowbit = len(bin(seed_)[2:].split("0")[-1])
        lastq[:] = lastq ^ samples[:, lowbit]

    #Calculate the new components of QUASI.
    quasi = numpy.empty((dim, order))
    for idx in range(order):
        lowbit = len(bin(seed+idx)[2:].split("0")[-1])
        quasi[:, idx] = lastq * recipd
        lastq[:] = lastq ^ samples[:, lowbit]

    return quasi



"""Collection of variance reduction techniques."""
from .generator import generate_samples

from .sequences import *  # pylint: disable=wildcard-import
from .latin_hypercube import create_latin_hypercube_samples
from .antithetic import create_antithetic_variates



"""Gauss-Legendre quadrature rule."""
try:
    from functools import lru_cache
except ImportError:  # pragma: no cover
    from functools32 import lru_cache

import numpy
import chaospy

from .hypercube import hypercube_quadrature


def legendre(order, lower=-1., upper=1., physicist=False):
    """
    Gauss-Legendre quadrature rule.

    Compute the sample points and weights for Gauss-Legendre quadrature. The
    sample points are the roots of the N-th degree Legendre polynomial. These
    sample points and weights correctly integrate polynomials of degree
    :math:`2N-1` or less over the interval ``[lower, upper]``.

    Gaussian quadrature come in two variants: physicist and probabilist. For
    Gauss-Legendre physicist means a weight function constant 1 and weights sum
    to ``upper-lower``, and probabilist means weight function constant
    ``1/(upper-lower)`` while weights sum to 1.

    Args:
        order (int):
            The quadrature order.
        lower (float):
            Lower bound for the integration interval.
        upper (float):
            Upper bound for the integration interval.
        physicist (bool):
            Use physicist weights instead of probabilist.

    Returns:
        abscissas (numpy.ndarray):
            The ``order+1`` quadrature points for where to evaluate the model
            function with.
        weights (numpy.ndarray):
            The quadrature weights associated with each abscissas.

    Examples:
        >>> abscissas, weights = chaospy.quadrature.legendre(2)
        >>> abscissas
        array([[-0.77459667,  0.        ,  0.77459667]])
        >>> weights
        array([0.27777778, 0.44444444, 0.27777778])

    See also:
        :func:`chaospy.quadrature.gaussian`
        :func:`chaospy.quadrature.legendre_proxy`

    """
    abscissas, weights = hypercube_quadrature(
        legendre_simple,
        order=int(order),
        domain=(float(lower), float(upper)),
    )
    weights = weights if physicist else weights/(upper-lower)
    return abscissas, weights


def legendre_proxy(
        order,
        domain=(0, 1),
        segments=1,
):
    r"""
    Generate proxy abscissas and weights from Legendre quadrature.

    Legendre provides optimal abscissas :math:`X_i` and weights :math:`W_i` to
    solve the integration problem:

    .. math::

        \int f(x) dx \approx \sum W_i f(X_i)

    over a function :math:`f`, where the probability density function :math:`p`
    is uniform.

    Since the weight function is constant, it can in principle be used to
    integrate any density function by considering it a part of the function. In
    other words:

    .. math::

        \int p(x) f(x) \approx \sum W_i p(X_i) f(X_i) = \sum W_i^' f(X_i)

    So when providing non-uniform distribution as `domain`, the weights will be
    adjusted with:

    .. math::

        W_i^' = W_i p(X_i)

    Bounds of the Legendre schemes is chosen to be the same as the distribution
    provided. This makes it a bad choice for unbound distributions.

    To get optimal abscissas and weights directly from a density, use
    :func:`chaospy.quadrature.gaussian` instead.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (:class:`chaospy.Distribution`, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        segments (int):
            Split intervals into steps subintervals and create a patched
            quadrature based on the segmented quadrature. Can not be lower than
            `order`. If 0 is provided, default to square root of `order`.
            Nested samples only appear when the number of segments are fixed.

    Returns:
        abscissas (numpy.ndarray):
            The quadrature points for where to evaluate the model function with
            ``abscissas.shape == (len(dist), N)`` where ``N`` is the number of
            samples.
        weights (numpy.ndarray):
            The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> abscissas, weights = chaospy.quadrature.legendre_proxy(3)
        >>> abscissas.round(4)
        array([[0.0694, 0.33  , 0.67  , 0.9306]])
        >>> weights.round(4)
        array([0.1739, 0.3261, 0.3261, 0.1739])
        >>> abscissas, weights = chaospy.quadrature.legendre_proxy(3, chaospy.Uniform(0, 1))
        >>> abscissas.round(4)
        array([[0.0694, 0.33  , 0.67  , 0.9306]])
        >>> weights.round(4)
        array([0.1739, 0.3261, 0.3261, 0.1739])
        >>> abscissas, weights = chaospy.quadrature.legendre_proxy(3, chaospy.Beta(2, 2))
        >>> abscissas.round(4)
        array([[0.0694, 0.33  , 0.67  , 0.9306]])
        >>> weights.round(4)
        array([0.0674, 0.4326, 0.4326, 0.0674])

    See also:
        :func:`chaospy.quadrature.legendre`,
        :func:`chaospy.quadrature.gaussian`

    """
    return hypercube_quadrature(
        legendre_simple,
        order=order,
        domain=domain,
        segments=segments,
    )


@lru_cache(None)
def legendre_simple(order):
    """
    Simple Legendre quadrature on the [0, 1] interval.

    Use :func:`chaospy.quadrature.legendre` instead.
    """
    coefficients = chaospy.construct_recurrence_coefficients(
        order=int(order), dist=chaospy.Uniform(-1, 1))
    [abscissas], [weights] = chaospy.coefficients_to_quadrature(coefficients)
    return abscissas*0.5+0.5, weights



r"""Collection of polynomial expansion constructors."""
import logging
from functools import wraps

from .chebyshev import chebyshev_1, chebyshev_2
from .cholesky import cholesky
from .frontend import generate_expansion
from .gegenbauer import gegenbauer
from .gram_schmidt import gram_schmidt
from .hermite import hermite
from .jacobi import jacobi
from .stieltjes import stieltjes
from .lagrange import lagrange
from .laguerre import laguerre
from .legendre import legendre

__all__ = ["generate_expansion"]


def expansion_deprecation_warning(name, func):

    @wraps(func)
    def wrapped(*args, **kwargs):
        """Function wrapper adds warnings."""
        logger = logging.getLogger(__name__)
        logger.warning("chaospy.%s name is to be deprecated; "
                       "Use chaospy.expansion.%s instead",
                       name, func.__name__)
        return func(*args, **kwargs)

    globals()[name] = wrapped
    __all__.append(name)


expansion_deprecation_warning("orth_ttr", stieltjes)
expansion_deprecation_warning("orth_chol", cholesky)
expansion_deprecation_warning("orth_gs", gram_schmidt)
expansion_deprecation_warning("lagrange_polynomial", lagrange)



r"""Collection of quadrature methods."""
import logging
from functools import wraps

from .frontend import generate_quadrature
from .sparse_grid import sparse_grid
from .utils import combine

from .chebyshev import chebyshev_1, chebyshev_2
from .clenshaw_curtis import clenshaw_curtis
from .discrete import discrete
from .fejer_1 import fejer_1
from .fejer_2 import fejer_2
from .gaussian import gaussian
from .genz_keister import (
    genz_keister_16, genz_keister_18, genz_keister_22, genz_keister_24)
from .gegenbauer import gegenbauer
from .grid import grid
from .hermite import hermite
from .jacobi import jacobi
from .kronrod import kronrod, kronrod_jacobi
from .laguerre import laguerre
from .legendre import legendre, legendre_proxy
from .leja import leja
from .lobatto import lobatto
from .newton_cotes import newton_cotes
from .patterson import patterson
from .radau import radau

__all__ = ["generate_quadrature", "sparse_grid", "combine"]


INTEGRATION_COLLECTION = {
    "clenshaw_curtis": clenshaw_curtis,
    "discrete": discrete,
    "fejer_1": fejer_1,
    "fejer_2": fejer_2,
    "gaussian": gaussian,
    "genz_keister_16": genz_keister_16,
    "genz_keister_18": genz_keister_18,
    "genz_keister_22": genz_keister_22,
    "genz_keister_24": genz_keister_24,
    "grid": grid,
    "kronrod": kronrod,
    "legendre": legendre_proxy,
    "leja": leja,
    "lobatto": lobatto,
    "newton_cotes": newton_cotes,
    "patterson": patterson,
    "radau": radau,
}


def quadrature_deprecation_warning(name, func):
    """Announce deprecation warning for quad-func."""
    quad_name = "quad_%s" % name

    @wraps(func)
    def wrapped(*args, **kwargs):
        """Function wrapper adds warnings."""
        logger = logging.getLogger(__name__)
        logger.warning("chaospy.%s name is to be deprecated; "
                       "Use chaospy.quadrature.%s instead",
                       quad_name, func.__name__)
        return func(*args, **kwargs)

    globals()[quad_name] = wrapped
    __all__.append(quad_name)

quadrature_deprecation_warning("clenshaw_curtis", clenshaw_curtis)
quadrature_deprecation_warning("discrete", discrete)
quadrature_deprecation_warning("fejer", fejer_2)
quadrature_deprecation_warning("grid", grid)
quadrature_deprecation_warning("gaussian", gaussian)
quadrature_deprecation_warning("newton_cotes", newton_cotes)
quadrature_deprecation_warning("leja", leja)
quadrature_deprecation_warning("gauss_legendre", legendre_proxy)
quadrature_deprecation_warning("gauss_kronrod", kronrod)
quadrature_deprecation_warning("gauss_lobatto", lobatto)
quadrature_deprecation_warning("gauss_patterson", patterson)
quadrature_deprecation_warning("gauss_radau", radau)
quadrature_deprecation_warning("genz_keister", genz_keister_24)



"""Distribution utility functions."""
import sys
import os
import logging
from contextlib import wraps
from itertools import permutations

import numpy
import chaospy


def check_dependencies(distribution):
    """
    Check if the dependency structure is valid.

    Rosenblatt transformations, density calculations etc. assumes that the
    input and output of transformation is the same. It also assumes that there
    is a order defined in `distribution._rotation` so an decomposition on the
    form `p(x0), p(x1|x0), ...` is possible.

    This should be checked on function calls, not init, as it does not apply to
    e.g. moment and three-terms-recurrence calculations.

    Args:
        distribution:
            The distribution to verify is correctly set up.

    Raises:
        StochasticallyDependentError:
            If invalid dependency structure is present.

    Examples:
        >>> dist = chaospy.Uniform(0, 1)
        >>> chaospy.check_dependencies(dist)
        >>> chaospy.check_dependencies(chaospy.Normal(mu=dist))
        Traceback (most recent call last):
            ...
        chaospy.StochasticallyDependentError: \
Normal(mu=Uniform(), sigma=1) has dangling dependencies

    """
    current = set()
    for idx in distribution._rotation:
        length = len(current)
        current.update(distribution._dependencies[idx])
        if len(current) != length+1:
            raise chaospy.StochasticallyDependentError(
                "%s has dangling dependencies" % distribution)


def shares_dependencies(*distributions):
    """
    Check if a collection of distributions shares dependencies.

    Internal dependencies are ignored here. It is only intra-dependencies that
    are checked. Anything else than a distribution is ignored.

    Args:
        distributions (Sequence[chaospy.Distribution]):
            The distributions to check shares dependencies or not.

    Returns:
        True if there are dependencies between the provided distributions.

    Examples:
        >>> dist = chaospy.Normal()
        >>> chaospy.shares_dependencies(dist, 4)
        False
        >>> chaospy.shares_dependencies(dist, chaospy.Normal())
        False
        >>> chaospy.shares_dependencies(dist, chaospy.Normal(dist))
        True

    """
    distributions = [dist for dist in distributions
                     if isinstance(dist, chaospy.Distribution)]
    if len(distributions) == 1:
        return False
    dependencies = [{dep for deps in dist._dependencies for dep in deps}
                    for dist in distributions]
    for deps1, deps2 in permutations(dependencies, 2):
        if deps1.intersection(deps2):
            return True
    return False


def declare_dependencies(
        distribution,
        parameters,
        rotation=None,
        is_operator=False,
        dependency_type="iid",
        length=None,
        extra_parameters=None,
):
    """
    Convenience function for declaring distribution dependencies.

    Iterates through parameters to figure out what a distribution dependency
    structure should be.

    Args:
        distribution (chaospy.Distribution):
            The distributions to to declare dependencies for.
        parameters (Dict[str, Any]):
            The distribution parameters that should be included in the
            declaration.
        rotation (Optional[Sequence[int]]):
            The order of which the dependencies should be resolved.
            Automatically calculated if omitted.
        is_operator (bool):
            Operators do not themselves contain uncertainty, but only inherits
            from parameters and/or wrapped distribution.
        wrapper_dist (Optional[chaospy.Distribution]):
            Distributions that are thin-wrappers to some other distribution
            should inherent dependencies.
        extra_parameters(Optional[Dict[str, Any]]):
            Extra parameters that should be included in the declaration, but
            not considered a direct part of the distribution. Assumed to be
            pre-processed.

    Returns:
        dependencies (List[Set[int]]):
            Dependency reference numbers, one collection per dimension. Single
            element implies stochstically independent.
        parameters (Dict[str, Any]):
            Same as `parameters`, but updated to all conform to the same size.
        rotation (List[int]):
            Same as `rotation` if provided. If not, the automatically
            calculated one is returned.

    """
    extra_parameters = extra_parameters.copy() if extra_parameters else {}
    parameters = parameters.copy()
    for name, parameter in list(parameters.items()):
        if not isinstance(parameter, chaospy.Distribution):
            parameters[name] = numpy.atleast_1d(parameter)
    if length is None:
        if rotation is None:
            length = max([len(parameter) for parameter in extra_parameters.values()]+
                         [len(parameter) for parameter in parameters.values()]+[1])
        else:
            length = len(rotation)
    if rotation is None:
        rotation = numpy.arange(length, dtype=int)
    else:
        rotation = numpy.asarray(rotation)

    if is_operator:
        dependencies = [set() for _ in range(length)]
    else:
        dependencies = init_dependencies(distribution, rotation, dependency_type=dependency_type)

    for name, parameter in list(parameters.items()):
        if isinstance(parameter, chaospy.Distribution):
            if len(parameter) != length:
                raise chaospy.StochasticallyDependentError(
                    "dependencies must be same length as parent")
        else:
            parameters[name] = parameter*numpy.ones(length, dtype=int)
    for name, parameter in list(parameters.items())+list(extra_parameters.items()):
        if isinstance(parameter, chaospy.Distribution):
            for dep1, dep2 in zip(dependencies, parameter._dependencies):
                dep1.update(dep2)
    assert len(dependencies) == length
    return dependencies, parameters, rotation


DISTRIBUTION_IDENTIFIERS = {}

def init_dependencies(
        distribution,
        rotation,
        dependency_type="iid",
):
    """
    Declare stochastic dependency to an underlying random variable.

    To be used inside distribution initializers ``__init__`` to map out the
    dependency structure.

    Args:
        distribution (chaospy.Distribution):
            Distribution to declare dependency for.
        rotation (Sequence[int]):
            The order of which the dependencies should be resolved.
        dependency_type (str):
            The type of dependency structure to create. Choose from "iid"
            (independent) and "accumulate" (saturated dependencies).

    Returns:
        (List[Set[int]]):
            Unique integer identifiers that represents dependencies.

    Examples:
        >>> from chaospy.distributions.baseclass.utils import DISTRIBUTION_IDENTIFIERS
        >>> DISTRIBUTION_IDENTIFIERS.clear()
        >>> core = chaospy.Normal()
        >>> core._dependencies
        [{0}]
        >>> DISTRIBUTION_IDENTIFIERS
        {0: normal()}
        >>> distribution = chaospy.Iid(core, 2)
        >>> distribution._dependencies
        [{1}, {2}]
        >>> chaospy.init_dependencies(distribution, [0, 1])
        [{3}, {4}]
        >>> chaospy.init_dependencies(distribution, [0, 1], "accumulate")
        [{5}, {5, 6}]
        >>> chaospy.init_dependencies(distribution, [1, 0])
        [{8}, {7}]
        >>> DISTRIBUTION_IDENTIFIERS[8]
        Iid(Normal(mu=0, sigma=1), 2)

    """
    rotation = numpy.asarray(rotation)
    assert rotation.dtype == int and rotation.ndim == 1
    length = len(rotation)
    next_new_id = len(DISTRIBUTION_IDENTIFIERS)
    new_identifiers = numpy.arange(next_new_id, next_new_id+length, dtype=int)
    for idx in new_identifiers:
        DISTRIBUTION_IDENTIFIERS[idx] = distribution

    if dependency_type == "iid":
        dependencies = [{idx} for idx in new_identifiers[rotation]]

    elif dependency_type == "accumulate":
        accumulant = set()
        dependencies = [None]*length
        for idx in rotation:
            accumulant.add(new_identifiers[idx])
            dependencies[idx] = accumulant.copy()

    return dependencies


def format_repr_kwargs(**parameters):
    """
    Format arguments for REPR output.

    Remove arguments if value are the same as their defaults. But only if that
    applies to all arguments.

    Args:
        parameters (Tuple[Any, numpy.number]):
            Parameters to format. First value is the parameters and the second
            is the parameters default. Using None for the default means no
            default.

    Returns:
        (List[str]):
            Positional argument to be used for REPR output.

    Examples:
        >>> chaospy.format_repr_kwargs(a=(4, 3), b=(chaospy.Uniform(), 4))
        ['a=4', 'b=Uniform()']
        >>> chaospy.format_repr_kwargs(a=(4, 3), b=(4, 4))
        ['a=4', 'b=4']
        >>> chaospy.format_repr_kwargs(a=(3, 3), b=(4, 4))
        []

    """
    out = []

    defaults_only = True
    for name, (param, default) in list(parameters.items()):
        defaults_only &= (
            (isinstance(param, (int, float)) and param == default) or
            (param is default is None)
        )
        if isinstance(param, numpy.ndarray):
            parameters[name] = (param.tolist(), default)
    if defaults_only:
        return []
    return ["%s=%s" % (key, parameters[key][0]) for key in sorted(parameters)]



"""Gaussian Mixture Model."""
import numpy

from .gaussian import GaussianKDE


class GaussianMixture(GaussianKDE):
    """
    Gaussian Mixture Model.

    A Gaussian mixture model is a probabilistic model that assumes all the data
    points are generated from a mixture of a finite number of Gaussian
    distributions with unknown parameters. One can think of mixture models as
    generalizing K-means clustering to incorporate information about the
    covariance structure of the data as well as the centers of the latent
    Gaussians.

    Attributes:
        means:
            Sequence of means.
        covariances:
            Sequence of covariance matrices.
        weights:
            How much each sample is weighted. Either a scalar when the samples
            are equally weighted, or a vector with the same length as the
            number of mixed models.

    Examples:
        >>> means = ([0, 1], [1, 0])
        >>> covariances = ([[1, 0], [0, 1]], [[1, 0.5], [0.5, 1]])
        >>> distribution = GaussianMixture(means, covariances)
        >>> uloc = [[0, 0, 1, 1], [0, 1, 0, 1]]
        >>> distribution.pdf(uloc).round(4)
        array([0.0954, 0.092 , 0.1212, 0.0954])
        >>> distribution.fwd(uloc).round(4)
        array([[0.3293, 0.3293, 0.6707, 0.6707],
               [0.3699, 0.6731, 0.3711, 0.734 ]])
        >>> distribution.inv(uloc).round(4)
        array([[-8.9681, -8.9681,  8.0521,  8.0521],
               [-9.862 , 10.1977, -9.5929, 10.2982]])
        >>> distribution.mom([(0, 1, 1), (1, 0, 1)]).round(4)
        array([0.5 , 0.5 , 0.25])

    """

    @property
    def means(self):
        return self._samples.T

    @property
    def covariances(self):
        return numpy.swapaxes(self._covariance, 0, 2)

    def __init__(self, means, covariances, weights=None, rotation=None):
        """
        Args:
            means (numpy.ndarray):
                Sequence of mean values. With shape `(n_components, n_dim)`.
            covariances (numpy.ndarray):
                Sequence of covariance matrices.
                With shape `(n_components, n_dim, n_dim)`.
            weights (Optional[numpy.ndarray]):
                Weights of the samples. This must have the shape
                `(n_components,)`. If omitted, each sample is assumed to be
                equally weighted.

        """
        means = numpy.atleast_2d(numpy.transpose(means))
        n, m = means.shape

        covariances = numpy.asfarray(covariances)
        if covariances.ndim in (1, 2):
            covariances = numpy.broadcast_to(covariances.T, (n, n, m))
        else:
            covariances = numpy.swapaxes(covariances, 0, 2)
        assert covariances.shape == (n, n, m)
        super(GaussianMixture, self).__init__(
            samples=means, h_mat=covariances, weights=weights, rotation=rotation)



"""Smolyak sparse grid constructor."""
from collections import defaultdict
from itertools import product

import numpy
from scipy.special import comb

import numpoly
import chaospy


def sparse_grid(
        order,
        dist,
        growth=None,
        recurrence_algorithm="stieltjes",
        rule="gaussian",
        tolerance=1e-10,
        scaling=3,
        n_max=5000,
):
    """
    Smolyak sparse grid constructor.

    Args:
        order (int, numpy.ndarray):
            The order of the grid. If ``numpy.ndarray``, it overrides both
            ``dim`` and ``skew``.
        dist (chaospy.distributions.baseclass.Distribution):
            The distribution which density will be used as weight function.
        growth (bool, None):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples. Defaults to the same
            value as ``sparse`` if omitted.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights in
            case of Gaussian quadrature scheme. If omitted, ``analytical`` will
            be tried first, and ``stieltjes`` used if that fails.
        rule (str):
            Rule for generating abscissas and weights. Either done with
            quadrature rules, or with random samples with constant weights.
        tolerance (float):
            The allowed relative error in norm between two quadrature orders
            before method assumes convergence.
        scaling (float):
            A multiplier the adaptive order increases with for each step
            quadrature order is not converged. Use 0 to indicate unit
            increments.
        n_max (int):
            The allowed number of quadrature points to use in approximation.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            Abscissas and weights created from sparse grid rule. Flatten such
            that ``abscissas.shape == (len(dist), len(weights))``.

    Example:
        >>> distribution = chaospy.J(chaospy.Normal(0, 1), chaospy.Uniform(-1, 1))
        >>> abscissas, weights = chaospy.quadrature.sparse_grid(1, distribution)
        >>> abscissas.round(4)
        array([[-1.    ,  0.    ,  0.    ,  0.    ,  1.    ],
               [ 0.    , -0.5774,  0.    ,  0.5774,  0.    ]])
        >>> weights.round(4)
        array([ 0.5,  0.5, -1. ,  0.5,  0.5])
        >>> abscissas, weights = chaospy.quadrature.sparse_grid([2, 1], distribution)
        >>> abscissas.round(2)
        array([[-1.73, -1.  , -1.  , -1.  ,  0.  ,  1.  ,  1.  ,  1.  ,  1.73],
               [ 0.  , -0.58,  0.  ,  0.58,  0.  , -0.58,  0.  ,  0.58,  0.  ]])
        >>> weights.round(2)
        array([ 0.17,  0.25, -0.5 ,  0.25,  0.67,  0.25, -0.5 ,  0.25,  0.17])
    """
    orders = order*numpy.ones(len(dist), dtype=int)
    growth = True if growth is None else growth

    assert isinstance(dist, chaospy.Distribution), "dist must be chaospy.Distribution"
    dist = dist if isinstance(dist, (chaospy.J, chaospy.Iid)) else chaospy.J(dist)

    if isinstance(rule, str):
        rule = (rule,)*len(dist)

    x_lookup, w_lookup = _construct_lookup(
        orders=orders,
        dists=dist,
        growth=growth,
        recurrence_algorithm=recurrence_algorithm,
        rules=rule,
        tolerance=tolerance,
        scaling=scaling,
        n_max=n_max,
    )
    collection = _construct_collection(
        order, dist, x_lookup, w_lookup)

    abscissas = sorted(collection)
    weights = numpy.array([collection[key] for key in abscissas])
    abscissas = numpy.array(abscissas).T
    return abscissas, weights


def _construct_collection(
        orders,
        dist,
        x_lookup,
        w_lookup,
):
    """Create a collection of {abscissa: weight} key-value pairs."""
    order = numpy.min(orders)
    skew = orders-order

    # Indices and coefficients used in the calculations
    indices = numpoly.glexindex(
        order-len(dist)+1, order+1, dimensions=len(dist))
    coeffs = numpy.sum(indices, -1)
    coeffs = (2*((order-coeffs+1) % 2)-1)*comb(len(dist)-1, order-coeffs)

    collection = defaultdict(float)
    for bidx, coeff in zip(indices+skew, coeffs.tolist()):
        abscissas = [value[idx] for idx, value in zip(bidx, x_lookup)]
        weights = [value[idx] for idx, value in zip(bidx, w_lookup)]
        for abscissa, weight in zip(product(*abscissas), product(*weights)):
            collection[abscissa] += numpy.prod(weight)*coeff

    return collection


def _construct_lookup(
        orders,
        dists,
        growth,
        recurrence_algorithm,
        rules,
        tolerance,
        scaling,
        n_max,
):
    """
    Create abscissas and weights look-up table so values do not need to be
    re-calculatated on the fly.
    """
    x_lookup = []
    w_lookup = []
    for max_order, dist, rule in zip(orders, dists, rules):
        x_lookup.append([])
        w_lookup.append([])
        for order in range(max_order+1):
            (abscissas,), weights = chaospy.generate_quadrature(
                order=order,
                dist=dist,
                growth=growth,
                recurrence_algorithm=recurrence_algorithm,
                rule=rule,
                tolerance=tolerance,
                scaling=scaling,
                n_max=n_max,
            )
            x_lookup[-1].append(abscissas)
            w_lookup[-1].append(weights)
    return x_lookup, w_lookup



"""Spearman's correlation coefficient."""
import numpy
from scipy.stats import spearmanr

import numpoly
import chaospy


def Spearman(poly, dist=None, sample=10000, retall=False, **kws):
    """
    Calculate Spearman's rank-order correlation coefficient.

    Args:
        poly (numpoly.ndpoly):
            Polynomial of interest.
        dist (Distribution):
            Defines the space where correlation is taken.
        sample (int):
            Number of samples used in estimation.
        retall (bool):
            If true, return p-value as well.

    Returns:
        (float, numpy.ndarray):
            Correlation output ``rho``. Of type float if two-dimensional problem.
            Correleation matrix if larger.
        (float, numpy.ndarray):
            The two-sided p-value for a hypothesis test whose null hypothesis
            is that two sets of data are uncorrelated, has same dimension as
            ``rho``.

    Examples:
        >>> distribution = chaospy.MvNormal(
        ...     [3, 4], [[2, .5], [.5, 1]])
        >>> corr, pvalue = chaospy.Spearman(distribution, sample=50, retall=True)
        >>> corr.round(4)
        array([[1.   , 0.603],
               [0.603, 1.   ]])
        >>> pvalue.round(8)
        array([[0.00e+00, 3.58e-06],
               [3.58e-06, 0.00e+00]])

    """
    if isinstance(poly, chaospy.Distribution):
        poly, dist = numpoly.variable(len(poly)), poly
    else:
        poly = numpoly.polynomial(poly)
    samples = dist.sample(sample, **kws)
    corr = numpy.eye(len(poly))
    pval = numpy.zeros((len(poly), len(poly)))
    evals = poly.ravel()(*samples)
    assert len(poly) == len(evals)
    for idx in range(len(poly)):
        for idy in range(idx+1, len(poly)):
            if idx == idy:
                pass
            spear = spearmanr(evals[idx], evals[idy])
            pval[idx, idy] = pval[idy, idx] = spear.pvalue
            corr[idx, idy] = corr[idy, idx] = spear.correlation
    if retall:
        return corr, pval
    return corr



"""Create samples from the Halton sequence."""
import numpy

from .van_der_corput import create_van_der_corput_samples
from .primes import create_primes


def create_halton_samples(order, dim=1, burnin=-1, primes=()):
    """
    Create samples from the Halton sequence.

    In statistics, Halton sequences are sequences used to generate points in
    space for numerical methods such as Monte Carlo simulations. Although these
    sequences are deterministic, they are of low discrepancy, that is, appear
    to be random for many purposes. They were first introduced in 1960 and are
    an example of a quasi-random number sequence. They generalise the
    one-dimensional van der Corput sequences. For ``dim == 1`` the sequence
    falls back to Van Der Corput sequence.

    Args:
        order (int):
            The order of the Halton sequence. Defines the number of samples.
        dim (int):
            The number of dimensions in the Halton sequence.
        burnin (int):
            Skip the first ``burnin`` samples. If negative, the maximum of
            ``primes`` is used.
        primes (tuple):
            The (non-)prime base to calculate values along each axis. If
            empty, growing prime values starting from 2 will be used.

    Returns:
        (numpy.ndarray):
            Halton sequence with ``shape == (dim, order)``.

    Examples:
        >>> distribution = chaospy.J(chaospy.Uniform(0, 1), chaospy.Uniform(0, 1))
        >>> samples = distribution.sample(3, rule="halton")
        >>> samples.round(4)
        array([[0.125 , 0.625 , 0.375 ],
               [0.4444, 0.7778, 0.2222]])
        >>> samples = distribution.sample(4, rule="halton")
        >>> samples.round(4)
        array([[0.125 , 0.625 , 0.375 , 0.875 ],
               [0.4444, 0.7778, 0.2222, 0.5556]])

    """
    primes = list(primes)
    if not primes:
        prime_order = 10*dim
        while len(primes) < dim:
            primes = create_primes(prime_order)
            prime_order *= 2
    primes = primes[:dim]
    assert len(primes) == dim, "not enough primes"

    if burnin < 0:
        burnin = max(primes)

    out = numpy.empty((dim, order))
    indices = [idx+burnin for idx in range(order)]
    for dim_ in range(dim):
        out[dim_] = create_van_der_corput_samples(
            indices, number_base=primes[dim_])
    return out



# -*- coding: utf-8 -*-
"""Generate the quadrature abscissas and weights in Fejer quadrature."""
import numpy
import chaospy

from .hypercube import hypercube_quadrature
from .clenshaw_curtis import clenshaw_curtis_simple


def fejer_2(order, domain=(0, 1), growth=False, segments=1):
    """
    Generate the quadrature abscissas and weights in Fejér type II quadrature.

    Fejér proposed two quadrature rules very similar to
    :func:`chaospy.quadrature.clenshaw_curtis`. The only difference is that the
    endpoints are removed. That is, Fejér only used the interior extrema of the
    Chebyshev polynomials, i.e. the true stationary points. This makes this a
    better method for performing quadrature on infinite intervals, as the
    evaluation does not contain endpoint values.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.Distribution, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.
        segments (int):
            Split intervals into N subintervals and create a patched
            quadrature based on the segmented quadrature. Can not be lower than
            `order`. If 0 is provided, default to square root of `order`.
            Nested samples only exist when the number of segments are fixed.

    Returns:
        abscissas (numpy.ndarray):
            The quadrature points for where to evaluate the model function
            with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
            number of samples.
        weights (numpy.ndarray):
            The quadrature weights with ``weights.shape == (N,)``.

    Notes:
        Implemented as proposed by Waldvogel :cite:`waldvogel_fast_2006`.

    Example:
        >>> abscissas, weights = chaospy.quadrature.fejer_2(3, (0, 1))
        >>> abscissas.round(4)
        array([[0.0955, 0.3455, 0.6545, 0.9045]])
        >>> weights.round(4)
        array([0.1804, 0.2996, 0.2996, 0.1804])

    See also:
        :func:`chaospy.quadrature.gaussian`
        :func:`chaospy.quadrature.clenshaw_curtis`
        :func:`chaospy.quadrature.fejer_1`

    """
    order = numpy.asarray(order)
    order = numpy.where(growth, numpy.where(order > 0, 2**(order+1)-2, 0), order)
    return hypercube_quadrature(
        quad_func=fejer_2_simple,
        order=order,
        domain=domain,
        segments=segments,
    )


def fejer_2_simple(order):
    """
    Backend for Fejer type II quadrature.

    Same as Clenshaw-Curtis, but with the end nodes removed.
    """
    abscissas, weights = clenshaw_curtis_simple(order+2)
    return abscissas[1:-1], weights[1:-1]



"""Chebyshev-Gauss quadrature rule of the first kind."""
import numpy
import chaospy

from .hypercube import hypercube_quadrature


def chebyshev_1(order, lower=-1, upper=1, physicist=False):
    r"""
    Chebyshev-Gauss quadrature rule of the first kind.

    Compute the sample points and weights for Chebyshev-Gauss quadrature. The
    sample points are the roots of the nth degree Chebyshev polynomial. These
    sample points and weights correctly integrate polynomials of degree
    :math:`2N-1` or less.

    Gaussian quadrature come in two variants: physicist and probabilist. For
    first order Chebyshev-Gauss physicist means a weight function
    :math:`1/\sqrt{1-x^2}` and weights that sum to :math`1/2`, and probabilist
    means a weight function is :math:`1/\sqrt{x (1-x)}` and sum to 1.

    Args:
        order (int):
            The quadrature order.
        lower (float):
            Lower bound for the integration interval.
        upper (float):
            Upper bound for the integration interval.
        physicist (bool):
            Use physicist weights instead of probabilist.

    Returns:
        abscissas (numpy.ndarray):
            The ``order+1`` quadrature points for where to evaluate the model
            function with.
        weights (numpy.ndarray):
            The quadrature weights associated with each abscissas.

    Examples:
        >>> abscissas, weights = chaospy.quadrature.chebyshev_1(3)
        >>> abscissas
        array([[-0.92387953, -0.38268343,  0.38268343,  0.92387953]])
        >>> weights
        array([0.25, 0.25, 0.25, 0.25])

    See also:
        :func:`chaospy.quadrature.chebyshev_2`
        :func:`chaospy.quadrature.gaussian`

    """
    order = int(order)
    coefficients = chaospy.construct_recurrence_coefficients(
        order=order, dist=chaospy.Beta(0.5, 0.5, lower, upper))
    [abscissas], [weights] = chaospy.coefficients_to_quadrature(coefficients)
    weights *= 0.5 if physicist else 1
    return abscissas[numpy.newaxis], weights



def chebyshev_2(order, lower=-1, upper=1, physicist=False):
    r"""
    Chebyshev-Gauss quadrature rule of the second kind.

    Compute the sample points and weights for Chebyshev-Gauss quadrature. The
    sample points are the roots of the nth degree Chebyshev polynomial. These
    sample points and weights correctly integrate polynomials of degree
    :math:`2N-1` or less.

    Gaussian quadrature come in two variants: physicist and probabilist. For
    second order Chebyshev-Gauss physicist means a weight function
    :math:`\sqrt{1-x^2}` and weights that sum to :math`2`, and probabilist
    means a weight function is :math:`\sqrt{x (1-x)}` and sum to 1.

    Args:
        order (int):
            The quadrature order.
        lower (float):
            Lower bound for the integration interval.
        upper (float):
            Upper bound for the integration interval.
        physicist (bool):
            Use physicist weights instead of probabilist.

    Returns:
        abscissas (numpy.ndarray):
            The ``order+1`` quadrature points for where to evaluate the model
            function with.
        weights (numpy.ndarray):
            The quadrature weights associated with each abscissas.

    Examples:
        >>> abscissas, weights = chaospy.quadrature.chebyshev_2(3)
        >>> abscissas
        array([[-0.80901699, -0.30901699,  0.30901699,  0.80901699]])
        >>> weights
        array([0.1381966, 0.3618034, 0.3618034, 0.1381966])

    See also:
        :func:`chaospy.quadrature.chebyshev_1`
        :func:`chaospy.quadrature.gaussian`

    """
    order = int(order)
    coefficients = chaospy.construct_recurrence_coefficients(
        order=order, dist=chaospy.Beta(1.5, 1.5, lower, upper))
    [abscissas], [weights] = chaospy.coefficients_to_quadrature(coefficients)
    weights *= 2 if physicist else 1
    return abscissas[numpy.newaxis], weights



"""Additive recursion sequence."""
import numpy


def generalized_golden_ratio(dim):
    """
    Using nested radical formula to calculate generalized golden ratio.

    Args:
        dim (int):
            The number of dimension the ratio is to be used in.

    Returns:
        (float):
            The generalize golden ratio for dimension `dim`.

    Examples:
        >>> generalized_golden_ratio(1)
        1.618033988749895
        >>> generalized_golden_ratio(2)
        1.324717957244746
        >>> generalized_golden_ratio(100)
        1.0069208854344955
    """
    out = 1.7
    out_ = 1.
    while out != out_:
        out, out_ = out_, (1+out)**(1./(dim+1))
    return out_


def create_additive_recursion_samples(order, dim=1, seed=0.5, alpha=None):
    """
    Create samples from the additive recursion sequence.

    Args:
        order (int):
            The number of samples to produce.
        dim (int):
            The number of dimensions in the sequence.
        seed (float):
            Random seed to be used in the sequence.
        alpha (Optional[Sequence[float]]):
            The (irrational) numbers used to create additive recursion.
            If omitted, inverse of generalized golde-ratio values will be used.

    Returns:
        (numpy.ndarray):
            Samples from the additive recursion sequence,
            with shape `(dim, order)`.

    Examples:
        >>> chaospy.create_additive_recursion_samples(5, 1).round(4)
        array([[0.118 , 0.7361, 0.3541, 0.9721, 0.5902]])
        >>> chaospy.create_additive_recursion_samples(5, 2).round(4)
        array([[0.2549, 0.0098, 0.7646, 0.5195, 0.2744],
               [0.0698, 0.6397, 0.2095, 0.7794, 0.3492]])
        >>> chaospy.create_additive_recursion_samples(5, 3).round(4)
        array([[0.3192, 0.1383, 0.9575, 0.7767, 0.5959],
               [0.171 , 0.8421, 0.5131, 0.1842, 0.8552],
               [0.0497, 0.5994, 0.1491, 0.6988, 0.2485]])

    """
    assert isinstance(dim, int) and dim > 0
    assert 0 <= seed < 1
    if alpha is None:
        phi = generalized_golden_ratio(dim)
        alpha = (1./phi)**numpy.arange(1, dim+1) % 1
    assert isinstance(alpha, numpy.ndarray)
    assert alpha.shape == (dim,)
    return (seed+numpy.outer(alpha, numpy.arange(1, order+1))) % 1



import numpy
import chaospy


def jacobi(
        order,
        alpha,
        beta,
        lower=-1,
        upper=1,
        physicist=False,
        normed=False,
        retall=False,
):
    """
    Jacobi polynomial expansion.

    Examples:
        >>> polynomials, norms = chaospy.expansion.jacobi(4, 0.5, 0.5, retall=True)
        >>> polynomials
        polynomial([1.0, q0, q0**2-0.5, q0**3-0.75*q0, q0**4-q0**2+0.125])
        >>> norms
        array([1.       , 0.5      , 0.125    , 0.03125  , 0.0078125])
        >>> chaospy.expansion.jacobi(3, 0.5, 0.5,  physicist=True).round(4)
        polynomial([1.0, 1.5*q0, 2.5*q0**2-0.8333, 4.375*q0**3-2.1146*q0])
        >>> chaospy.expansion.jacobi(3, 1.5, 0.5, normed=True)
        polynomial([1.0, 2.0*q0, 4.0*q0**2-1.0, 8.0*q0**3-4.0*q0])

    """
    multiplier = 1
    if physicist:
        multiplier = numpy.arange(1, order+1)
        multiplier = ((2*multiplier+alpha+beta-1)*(2*multiplier+alpha+beta)/
                      (2*multiplier*(multiplier+alpha+beta)))
    _, [polynomials], [norms] = chaospy.recurrence.analytical_stieltjes(
        order, chaospy.Beta(alpha, beta, lower=lower, upper=upper), multiplier=multiplier)
    if normed:
        polynomials = chaospy.true_divide(polynomials, numpy.sqrt(norms))
        norms[:] = 1.
    return (polynomials, norms) if retall else polynomials



"""Generalized Gauss-Laguerre quadrature rule."""
import numpy
from scipy.special import gamma
import chaospy

from .hypercube import hypercube_quadrature


def laguerre(order, alpha=0., physicist=False):
    r"""
    Generalized Gauss-Laguerre quadrature rule.

    Compute the sample points and weights for Gauss-Laguerre quadrature. The
    sample points are the roots of the nth degree Laguerre polynomial. These
    sample points and weights correctly integrate polynomials of degree
    :math:`2N-1` or less.

    Gaussian quadrature come in two variants: physicist and probabilist. For
    Gauss-Laguerre physicist means a weight function :math:`x^\alpha e^{-x}`
    and weights that sum to :math`\Gamma(\alpha+1)`, and probabilist means a
    weight function is :math:`x^\alpha e^{-x}` and sum to 1.

    Args:
        order (int):
            The quadrature order.
        alpha (float):
            Shape parameter. Defaults to non-generalized Laguerre if 0.
        physicist (bool):
            Use physicist weights instead of probabilist.

    Returns:
        abscissas (numpy.ndarray):
            The ``order+1`` quadrature points for where to evaluate the model
            function with.
        weights (numpy.ndarray):
            The quadrature weights associated with each abscissas.

    Examples:
        >>> abscissas, weights = chaospy.quadrature.laguerre(2)
        >>> abscissas
        array([[0.41577456, 2.29428036, 6.28994508]])
        >>> weights
        array([0.71109301, 0.27851773, 0.01038926])

    See also:
        :func:`chaospy.quadrature.gaussian`

    """
    order = int(order)
    coefficients = chaospy.construct_recurrence_coefficients(
        order=order, dist=chaospy.Gamma(alpha+1))
    [abscissas], [weights] = chaospy.coefficients_to_quadrature(coefficients)
    weights *= gamma(alpha+1) if physicist else 1
    return abscissas[numpy.newaxis], weights



"""Triangle probability distribution."""
import numpy
from scipy import special, misc

from ..baseclass import SimpleDistribution, LowerUpperDistribution
from .beta import beta_

def _get_triang_param_lb_and_ub(name, value):
    try:
        lb = ub = float(value)
    except:
        if hasattr(value, 'lower') and hasattr(value, 'upper'):
            lb = value.lower.min()
            ub = value.upper.max()
        else:
            raise ValueError(
                "%s must be either a number or a distribution " % name
              + "with lower and upper bounds; not a '%s' " % type(value).__name__
              + "object"
            )
    return lb, ub

class triangle(SimpleDistribution):
    """Triangle probability distribution."""

    def __init__(self, a=.5):
        # assert numpy.all(a>=0) and numpy.all(a<=1)
        super(triangle, self).__init__(dict(a=a))

    def _pdf(self, xloc, a):
        return numpy.where(xloc < a, 2*xloc/a, 2*(1-xloc)/(1-a))

    def _cdf(self, xloc, a):
        return numpy.where(xloc < a, xloc**2/(a+(a == 0)),
                (2*xloc-xloc*xloc-a)/(1-a+(a == 1)))

    def _ppf(self, qloc, a):
        return numpy.where(qloc<a, numpy.sqrt(qloc*a), 1-numpy.sqrt(1-a-qloc*(1-a)))

    def _mom(self, k, a):
        a_ = a*(a!=1)
        out = 2*(1.-a_**(k+1))/((k+1)*(k+2)*(1-a_))
        return numpy.where(a==1, 2./(k+2), out)

    def _lower(self, a):
        return 0.

    def _upper(self, a):
        return 1.


class Triangle(LowerUpperDistribution):
    """
    Triangle Distribution.

    Must have lower <= midpoint <= upper.

    Args:
        lower (float, Distribution):
            Lower bound
        midpoint (float, Distribution):
            Location of the top
        upper (float, Distribution):
            Upper bound

    Examples:
        >>> distribution = chaospy.Triangle(-1, 0 ,1)
        >>> distribution
        Triangle(-1, 0, 1)
        >>> uloc = numpy.linspace(0, 1, 6)
        >>> uloc
        array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
        >>> xloc = distribution.inv(uloc)
        >>> xloc.round(3)
        array([-1.   , -0.368, -0.106,  0.106,  0.368,  1.   ])
        >>> numpy.allclose(distribution.fwd(xloc), uloc)
        True
        >>> distribution.pdf(xloc).round(3)
        array([0.   , 0.632, 0.894, 0.894, 0.632, 0.   ])
        >>> distribution.sample(4).round(3)
        array([ 0.168, -0.52 ,  0.685, -0.018])
        >>> distribution.mom(1).round(4)
        0.0
        >>> distribution = chaospy.Triangle(-1, 2 ,1)
        Traceback (most recent call last):
        ValueError: condition not satisfied: `lower <= midpoint <= upper`

    """

    def __init__(self, lower=-1, midpoint=0, upper=1):
        repr_args = [lower, midpoint, upper]
        _, lower_ub = _get_triang_param_lb_and_ub('lower', lower)
        mid_lb, mid_ub = _get_triang_param_lb_and_ub('midpoint', midpoint)
        upper_lb, _ = _get_triang_param_lb_and_ub('upper', upper)
        if not (lower_ub <= mid_lb  and mid_ub <= upper_lb):
            raise ValueError(
                "condition not satisfied: `lower <= midpoint <= upper`"
            )
        midpoint = (midpoint-lower)*1./(upper-lower)
        super(Triangle, self).__init__(
            dist=triangle(midpoint),
            lower=lower,
            upper=upper,
        )
        self._repr_args = repr_args



"""Discretized Stieltjes' method."""
import numpy
import numpoly
import chaospy


def stieltjes(
        order,
        dist,
        rule=None,
        tolerance=1e-16,
        scaling=3,
        n_max=5000,
):
    """
    Stieltjes' method.

    Tries to get recurrence coefficients using the distributions own
    TTR-method, but will fall back to a iterative method if missing.

    Args:
        order (int):
            The order create recurrence coefficients for.
        dist (chaospy.Distribution):
            The distribution to create recurrence coefficients with respect to.
        rule (str):
            The rule to use to create quadrature nodes and weights from.
        tolerance (float):
            The allowed relative error in norm between two quadrature orders
            before method assumes convergence.
        scaling (float):
            A multiplier the adaptive order increases with for each step
            quadrature order is not converged. Use 0 to indicate unit
            increments.
        n_max (int):
            The allowed number of quadrature points to use in approximation.

    Returns:
        (numpy.ndarray, numpy.ndarray, numpy.ndarray):
            coefficients:
                The recurrence coefficients created using the discretized
                Stieltjes' method, with ``shape == (2, D, order+1)``.
            polynomials:
                The orthogonal polynomial expansion created as a by-product of
                the algorithm.
            norms:
                The norm of each orthogonal polynomial. Roughly equivalent to
                ``chaospy.E(polynomials**2, dist)``, but more numerically
                stable than most alternatives.

    Examples:
        >>> dist = chaospy.J(chaospy.Uniform(0, 1), chaospy.Beta(3, 4))
        >>> (alpha, beta), orth, norms = chaospy.stieltjes(2, dist)
        >>> alpha.round(5)
        array([[0.5    , 0.5    , 0.5    ],
               [0.42857, 0.46032, 0.47475]])
        >>> beta.round(5)
        array([[1.     , 0.08333, 0.06667],
               [1.     , 0.03061, 0.04321]])
        >>> orth[:, 2].round(5)
        polynomial([q0**2-q0+0.16667, q1**2-0.88889*q1+0.16667])
        >>> norms.round(5)
        array([[1.     , 0.08333, 0.00556],
               [1.     , 0.03061, 0.00132]])

    """
    try:
        return analytical_stieltjes(order=order, dist=dist)
    except NotImplementedError:
        return discretized_stieltjes(
            order=order,
            dist=dist,
            rule=rule,
            tolerance=tolerance,
            scaling=scaling,
            n_max=n_max,
        )


def discretized_stieltjes(
        order,
        dist,
        rule=None,
        tolerance=1e-16,
        scaling=3,
        n_max=5000,
):
    """
    Discretized Stieltjes' method.

    Examples:
        >>> dist = chaospy.J(chaospy.Uniform(0, 1), chaospy.Beta(3, 4))
        >>> (alpha, beta), orth, norms = chaospy.discretized_stieltjes(2, dist)
        >>> alpha.round(5)
        array([[0.5    , 0.5    , 0.5    ],
               [0.42857, 0.46032, 0.47475]])
        >>> beta.round(5)
        array([[1.     , 0.08333, 0.06667],
               [1.     , 0.03061, 0.04321]])
        >>> orth[:, 2].round(5)
        polynomial([q0**2-q0+0.16667, q1**2-0.88889*q1+0.16667])
        >>> norms.round(5)
        array([[1.     , 0.08333, 0.00556],
               [1.     , 0.03061, 0.00132]])

    """
    if len(dist) > 1:
        assert not dist.stochastic_dependent
        coeffs, orths, norms = zip(*[discretized_stieltjes(
                order,
                dist_,
                rule=rule,
                tolerance=tolerance,
                scaling=scaling)
        for dist_ in dist])
        coeffs = numpy.dstack(coeffs).reshape(2, len(dist), order+1)
        variables = list(numpoly.variable(len(dist)))
        orths = [orths[idx](q0=variables[idx]) for idx in range(len(dist))]
        orths = numpoly.polynomial(orths).reshape(len(dist), order+1)
        norms = numpy.asfarray(norms).reshape(len(dist), order+1)
        return coeffs, orths, norms

    if rule is None:
        rule = "discrete" if dist.interpret_as_integer else "clenshaw_curtis"
    order_ = (2*order-1.)/scaling
    beta = beta_old = numpy.nan
    var = numpoly.variable()
    orths = [numpoly.polynomial(0.), numpoly.polynomial(1.)]+[None]*order
    norms = numpy.ones(order+2)
    coeffs = numpy.ones((2, order+1))

    while not numpy.all(numpy.abs(coeffs[1]-beta_old) < tolerance):

        beta_old = coeffs[1].copy()
        order_ = max(order_*scaling, order_+1)
        if order_ > n_max:
            break

        [abscissas], weights = chaospy.generate_quadrature(
            int(order_), dist, rule=rule, segments=0)
        inner = numpy.sum(abscissas*weights)
        for idx in range(order):
            coeffs[0, idx] = inner/norms[idx+1]
            coeffs[1, idx] = norms[idx+1]/norms[idx]
            orths[idx+2] = ((var-coeffs[0, idx])*orths[idx+1]-
                            orths[idx]*coeffs[1, idx])
            norms[idx+2] = numpy.sum(orths[idx+2](abscissas)**2*weights)
            inner = numpy.sum(abscissas*orths[idx+2](abscissas)**2*weights)
        coeffs[:, order] = (inner/norms[-1], norms[-1]/norms[-2])

    coeffs = coeffs.reshape(2, 1, order+1)
    orths = numpoly.polynomial(orths[1:]).reshape(1, order+1)
    norms = numpy.array(norms[1:]).reshape(1, order+1)
    return coeffs, orths, norms


def analytical_stieltjes(order, dist, multiplier=1):
    """Analytical Stieltjes' method"""
    dimensions = len(dist)
    mom_order = numpy.arange(order+1).repeat(dimensions)
    mom_order = mom_order.reshape(order+1, dimensions).T
    coeffs = dist.ttr(mom_order)
    coeffs[1, :, 0] = 1.
    orders = numpy.arange(order, dtype=int)
    multiplier, orders = numpy.broadcast_arrays(multiplier, orders)

    var = numpoly.variable(dimensions)
    orth = [numpy.zeros(dimensions), numpy.ones(dimensions)]
    for order_, multiplier_ in zip(orders, multiplier):
        orth.append(
            multiplier_*((var-coeffs[0, :, order_])*orth[-1]-coeffs[1, :, order_]*orth[-2]))
    orth = numpoly.polynomial(orth[1:]).T
    norms = numpy.cumprod(coeffs[1], 1)

    assert coeffs.shape == (2, dimensions, order+1)
    assert orth.shape == (len(dist), order+1)
    assert norms.shape == (len(dist), order+1)
    return coeffs, orth, norms



"""Gauss-Hermite quadrature rule."""
import numpy
import chaospy

from .hypercube import hypercube_quadrature


def hermite(order, mu=0., sigma=1., physicist=False):
    r"""
    Gauss-Hermite quadrature rule.

    Compute the sample points and weights for Gauss-Hermite quadrature. The
    sample points are the roots of the nth degree Hermite polynomial. These
    sample points and weights correctly integrate polynomials of degree
    :math:`2N-1` or less.

    Gaussian quadrature come in two variants: physicist and probabilist. For
    Gauss-Hermite physicist means a weight function :math:`e^{-x^2}` and
    weights that sum to :math`\sqrt(\pi)`, and probabilist means a weight
    function is :math:`e^{-x^2/2}` and sum to 1.

    Args:
        order (int):
            The quadrature order.
        mu (float):
            Non-centrality parameter.
        sigma (float):
            Scale parameter.
        physicist (bool):
            Use physicist weights instead of probabilist variant.

    Returns:
        abscissas (numpy.ndarray):
            The ``order+1`` quadrature points for where to evaluate the model
            function with.
        weights (numpy.ndarray):
            The quadrature weights associated with each abscissas.

    Examples:
        >>> abscissas, weights = chaospy.quadrature.hermite(3)
        >>> abscissas
        array([[-2.33441422, -0.74196378,  0.74196378,  2.33441422]])
        >>> weights
        array([0.04587585, 0.45412415, 0.45412415, 0.04587585])

    See also:
        :func:`chaospy.quadrature.gaussian`

    """
    order = int(order)
    sigma = float(sigma*2**-0.5 if physicist else sigma)
    coefficients = chaospy.construct_recurrence_coefficients(
        order=order, dist=chaospy.Normal(0, sigma))
    [abscissas], [weights] = chaospy.coefficients_to_quadrature(coefficients)
    weights = weights*numpy.pi**0.5 if physicist else weights
    if order%2 == 0:
        abscissas[len(abscissas)//2] = 0
    abscissas += mu
    return abscissas[numpy.newaxis], weights



from functools import partial

import numpy
import chaospy

from .utils import combine_quadrature


def hypercube_quadrature(
    quad_func,
    order,
    domain,
    segments=None,
    auto_scale=True,
):
    """
    Enhance simple 1-dimensional unit quadrature with extra features.

    These features include handling of:

    * Distribution as domain by embedding density into the weights
    * Scale to any intervals
    * Multivariate support
    * Repeat quadrature into segments

    Args:
        quad_func (Callable):
            Function that creates quadrature abscissas and weights. If
            ``auto_scale`` is true, the function should be on the form:
            ``abscissas, weights = quad_func(order)`` and be defined on the unit
            interval. Otherwise the call signature should be
            ``abscissas, weights = quad_func(order, lower, upper)`` and it
            should be defined on interval bound by ``lower`` and ``upper``.
        order (int, Sequence[int]):
            The quadrature order passed to the quadrature function.
        domain (Tuple[float, float], :class:`chaospy.Distribution`):
            Either interval on the format ``(lower, upper)`` or a distribution
            to integrate over. If the latter, weights are adjusted to
            incorporate the density at abscissas.
        segments (Optional[int], Sequence[float]):
            The number segments to split the interval on. If sequence is
            provided, use as segment edges instead.
        kwargs (Any):
            Extra keyword arguments passed to `quad_func`.

    Returns:
        Same as ``quad_func`` but adjusted to incorporate extra features.

    Examples:
        >>> def my_quad(order):
        ...     return (numpy.linspace(0, 1, order+1),
        ...             1./numpy.full(order+1, order+2))
        >>> my_quad(2)
        (array([0. , 0.5, 1. ]), array([0.25, 0.25, 0.25]))
        >>> hypercube_quadrature(my_quad, 2, domain=chaospy.Uniform(-1, 1))
        (array([[-1.,  0.,  1.]]), array([0.25, 0.25, 0.25]))
        >>> abscissas, weights = hypercube_quadrature(my_quad, (1, 1), domain=(0, 1))
        >>> abscissas
        array([[0., 0., 1., 1.],
               [0., 1., 0., 1.]])
        >>> weights.round(5)
        array([0.11111, 0.11111, 0.11111, 0.11111])

    """
    if segments is None:
        order, domain = align_arguments(order, domain)
        kwargs = dict(order=order)
    else:
        order, domain, segments = align_arguments(order, domain, segments)
        kwargs = dict(order=order, segments=segments)

    quad_func = partial(ensure_output, quad_func=quad_func)
    if auto_scale:
        if segments is not None:
            quad_func = partial(split_into_segments, quad_func=quad_func)
        quad_func = partial(scale_quadrature, quad_func=quad_func)
    quad_func = partial(univariate_to_multivariate, quad_func=quad_func)
    if isinstance(domain, chaospy.Distribution):
        quad_func = partial(
            distribution_to_domain, quad_func=quad_func, distribution=domain)
    else:
        quad_func = partial(
            quad_func, lower=numpy.asarray(domain[0]), upper=numpy.asarray(domain[1]))

    return quad_func(**kwargs)


def align_arguments(
    order,
    domain,
    segments=None,
):
    """
    Extract dimensions from input arguments and broadcast relevant parts.

    Args:
        order (int, Sequence[int]):
            The quadrature order passed to the quadrature function.
        domain (Tuple[float, float], :func:`chaospy.Distribution`):
            Either interval on the format ``(lower, upper)`` or a distribution
            to integrate over.
        segments (Optional[int], Sequence[float]):
            The number segments to split the interval on. If sequence is
            provided, use as segment edges instead.

    Examples:
        >>> order, domain, segments = align_arguments(1, chaospy.Uniform(0, 1), 1)
        >>> order, domain, segments
        (array([1]), Uniform(), array([1]))
        >>> distribution = chaospy.Iid(chaospy.Uniform(0, 1), 2)
        >>> order, domain = align_arguments(1, distribution)
        >>> order, domain
        (array([1, 1]), Iid(Uniform(), 2))

    """
    args = [numpy.asarray(order)]
    if isinstance(domain, chaospy.Distribution):
        args += [numpy.zeros(len(domain))]
    else:
        args += list(domain)
    if segments is not None:
        segments = numpy.atleast_1d(segments)
        assert segments.ndim <= 2
        if segments.ndim == 2:
            args.append(segments[:, 0])
        else:
            args.append(segments)
    args = numpy.broadcast_arrays(*args)

    output = [args.pop(0)]
    if not isinstance(domain, chaospy.Distribution):
        output += [(args.pop(0), args.pop(0))]
    else:
        output += [domain]
    if segments is not None:
        if segments.ndim == 2:
            segments = numpy.broadcast_arrays(segments, order)[0].T
        else:
            segments = args.pop(-1)
        output += [segments]
    return tuple(output)


def ensure_output(quad_func, **kwargs):
    """
    Converts arrays to python native types and ensure quadrature output sizes.

    Args:
        quad_func (Callable):
            Function that creates quadrature abscissas and weights.
        kwargs (Any):
            Extra keyword arguments passed to `quad_func`.

    Returns:
        Same as ``quad_func(order, **kwargs)`` except numpy elements in
        ``kwargs`` is replaced with Python native counterparts and
        ``abscissas`` is ensured to be at least 2-dimensional.

    Examples:
        >>> def my_quad(order):
        ...     return (numpy.linspace(0, 1, order+1),
        ...             1./numpy.full(order+1, order+2))
        >>> my_quad(2)
        (array([0. , 0.5, 1. ]), array([0.25, 0.25, 0.25]))
        >>> ensure_output(my_quad, order=numpy.array([2]))
        (array([[0. , 0.5, 1. ]]), array([0.25, 0.25, 0.25]))

    """
    kwargs = {key: (value.item() if isinstance(value, numpy.ndarray) else value)
              for key, value in kwargs.items()}
    abscissas, weights = quad_func(**kwargs)
    abscissas = numpy.atleast_2d(abscissas)
    assert abscissas.ndim == 2
    assert weights.ndim == 1
    assert abscissas.shape[-1] == len(weights)
    return abscissas, weights


def univariate_to_multivariate(quad_func, **kwargs):
    """
    Turn a univariate quadrature rule into a multivariate rule.

    The one-dimensional quadrature functions are combined into a multivariate
    through tensor-product. The dimensionality is inferred from the keyword
    arguments. Weights are adjusted to correspond to the multivariate scheme.

    Args:
        quad_func (Callable):
            Function that creates quadrature abscissas and weights on the unit
            interval.
        kwargs (Any):
            Keyword arguments passed to `quad_func`. If numerical value is
            provided, it is used to infer the dimensions of the multivariate
            output. Non-numerical values are passed as is.

    Returns:
        Same as ``quad_func(order, **kwargs)`` except with multivariate
        supported.

    Examples:
        >>> def my_quad(order):
        ...     return (numpy.linspace(0, 1, order+1)[numpy.newaxis],
        ...             1./numpy.full(order+1, order+2))
        >>> my_quad(1)
        (array([[0., 1.]]), array([0.33333333, 0.33333333]))
        >>> my_quad(2)
        (array([[0. , 0.5, 1. ]]), array([0.25, 0.25, 0.25]))
        >>> abscissas, weights = univariate_to_multivariate(
        ...     my_quad, order=numpy.array([1, 2, 1]))
        >>> abscissas
        array([[0. , 0. , 0. , 0. , 0. , 0. , 1. , 1. , 1. , 1. , 1. , 1. ],
               [0. , 0. , 0.5, 0.5, 1. , 1. , 0. , 0. , 0.5, 0.5, 1. , 1. ],
               [0. , 1. , 0. , 1. , 0. , 1. , 0. , 1. , 0. , 1. , 0. , 1. ]])
        >>> weights.round(6)
        array([0.027778, 0.027778, 0.027778, 0.027778, 0.027778, 0.027778,
               0.027778, 0.027778, 0.027778, 0.027778, 0.027778, 0.027778])
        >>> univariate_to_multivariate(my_quad, order=numpy.array([1]))
        (array([[0., 1.]]), array([0.33333333, 0.33333333]))

    """
    sizables = {key: value for key, value in kwargs.items()
                if isinstance(value, (int, float, numpy.ndarray))}
    sizables["_"] = numpy.zeros(len(kwargs.get("domain", [0])))
    nonsizables = {key: value for key, value in kwargs.items()
                   if not isinstance(value, (int, float, numpy.ndarray))}
    keys = list(sizables)
    args = numpy.broadcast_arrays(*[sizables[key] for key in keys])
    assert args[0].ndim == 1
    sizables = {key: value for key, value in zip(keys, args)}
    del sizables["_"]

    results = []
    for idx in range(args[0].size):
        sizable = kwargs.copy()
        sizable.update({key: value[idx].item()
                        for key, value in sizables.items()})
        abscissas, weights = quad_func(**sizable)
        results.append((abscissas.ravel(), weights))

    abscissas, weights = zip(*results)
    return combine_quadrature(abscissas, weights)


def distribution_to_domain(quad_func, distribution, **kwargs):
    """
    Integrate over a distribution domain.

    Adjust weights to account for probability density.

    Args:
        quad_func (Callable):
            Function that creates quadrature abscissas and weights. Must accept
            the arguments ``lower`` and ``upper`` to define the interval it is
            integrating over.
        distribution (:class:`chaospy.Distribution`):
            Distribution to adjust quadrature scheme to.
        kwargs (Any):
            Extra keyword arguments passed to `quad_func`. Can not include the
            arguments ``lower`` and ``upper`` as they are taken from
            ``distribution``.

    Returns:
        Same as ``quad_func(order, **kwargs)`` except arguments ``lower`` and
        ``upper`` are now replaced with a new ``distribution`` argument.

    Examples:
        >>> def my_quad(lower=0, upper=1):
        ...     return (numpy.linspace(lower, upper, 5).reshape(1, -1)[:, 1:-1],
        ...             1./numpy.full(3, 4))
        >>> my_quad()
        (array([[0.25, 0.5 , 0.75]]), array([0.25, 0.25, 0.25]))
        >>> distribution_to_domain(my_quad, chaospy.Uniform(-1, 1))
        (array([[-0.5,  0. ,  0.5]]), array([0.125, 0.125, 0.125]))
        >>> distribution_to_domain(my_quad, chaospy.Beta(2, 2))
        (array([[0.25, 0.5 , 0.75]]), array([0.225, 0.3  , 0.225]))
        >>> distribution_to_domain(my_quad, chaospy.Exponential(1))  # doctest: +NORMALIZE_WHITESPACE
        (array([[ 8.05924772, 16.11849545, 24.17774317]]),
         array([2.32578431e-02, 7.35330570e-06, 2.32485465e-09]))

    """
    assert isinstance(distribution, chaospy.Distribution)
    assert "lower" not in kwargs
    assert "upper" not in kwargs
    lower = distribution.lower
    upper = distribution.upper
    abscissas, weights = quad_func(lower=lower, upper=upper, **kwargs)

    # Sometimes edge samples (inside the distribution domain) falls out again from simple
    # rounding errors. Edge samples needs to be adjusted.
    eps = 1e-14*(distribution.upper-distribution.lower)
    abscissas_ = numpy.clip(abscissas.T, distribution.lower+eps, distribution.upper-eps).T
    weights_ = weights*distribution.pdf(abscissas_).ravel()
    weights = weights_*numpy.sum(weights)/(numpy.sum(weights_)*numpy.prod(upper-lower))
    return abscissas, weights


def split_into_segments(quad_func, order, segments, **kwargs):
    """
    Split a quadrature rule on ta unit interval to multiple segments.

    If both quadrature function includes the abscissas endpoints 0 and 1, then
    the endpoints of each subsequent interval is collapsed and their weights
    added together. This to avoid nodes from being repeated.

    Args:
        quad_func (Callable):
            Function that creates quadrature abscissas and weights on the unit
            interval.
        order (int):
            The quadrature order passed to the quadrature function.
        segments (int, Sequence[float]):
            The number segments to split the interval on. If sequence is
            provided, use as segment edges instead.
        kwargs (Any):
            Extra keyword arguments passed to `quad_func`.

    Returns:
        Same as ``quad_func(order, **kwargs)`` except segmented into
        subintervals.

    Examples:
        >>> def my_quad(order):
        ...     return (numpy.linspace(0, 1, order+1)[numpy.newaxis],
        ...             1./numpy.full(order+1, order+2))
        >>> my_quad(2)
        (array([[0. , 0.5, 1. ]]), array([0.25, 0.25, 0.25]))
        >>> split_into_segments(my_quad, 4, segments=2)  # doctest: +NORMALIZE_WHITESPACE
        (array([[0.  , 0.25, 0.5 , 0.75, 1.  ]]),
         array([0.125, 0.125, 0.25 , 0.125, 0.125]))
        >>> split_into_segments(my_quad, 4, segments=3)  # doctest: +NORMALIZE_WHITESPACE
        (array([[0.        , 0.16666667, 0.33333333, 0.66666667, 1.        ]]),
         array([0.08333333, 0.08333333, 0.19444444, 0.22222222, 0.11111111]))
        >>> split_into_segments(my_quad, 4, segments=[0.2, 0.8])  # doctest: +NORMALIZE_WHITESPACE
        (array([[0. , 0.1, 0.2, 0.5, 0.8, 0.9, 1. ]]),
         array([0.05, 0.05, 0.2 , 0.15, 0.2 , 0.05, 0.05]))

    """
    segments = numpy.array(segments)
    if segments.size == 1:
        segments = int(segments)
        if segments == 1 or order <= 2:
            return quad_func(order=order, **kwargs)
        if not segments:
            segments = int(numpy.sqrt(order))
        assert segments < order, "few samples to distribute than intervals"
        nodes = numpy.linspace(0, 1, segments+1)
    else:
        nodes, segments = numpy.hstack([[0], segments, [1]]), len(segments)

    abscissas = []
    weights = []
    for idx, (lower, upper) in enumerate(zip(nodes[:-1], nodes[1:])):

        assert lower < upper
        order_ = order//segments + (idx < (order%segments))
        abscissa, weight = quad_func(order=order_, **kwargs)
        weight = weight*(upper-lower)
        abscissa = (abscissa.T*(upper-lower)+lower).T
        if abscissas and numpy.allclose(abscissas[-1][:, -1], lower):
            weights[-1][-1] += weight[0]
            abscissa = abscissa[:, 1:]
            weight = weight[1:]
        abscissas.append(abscissa)
        weights.append(weight)

    abscissas = numpy.hstack(abscissas)
    weights = numpy.hstack(weights)
    assert abscissas.shape == (1, len(weights))
    return abscissas, weights


def scale_quadrature(quad_func, order, lower, upper, **kwargs):
    """
    Scale quadrature rule designed for unit interval to an arbitrary interval.

    Args:
        quad_func (Callable):
            Function that creates quadrature abscissas and weights on the unit
            interval.
        order (int):
            The quadrature order passed to the quadrature function.
        lower (float):
            The new lower limit for the quadrature function.
        upper (float):
            The new upper limit for the quadrature function.
        kwargs (Any):
            Extra keyword arguments passed to `quad_func`.

    Returns:
        Same as ``quad_func(order, **kwargs)`` except scaled to a new interval.


    Examples:
        >>> def my_quad(order):
        ...     return (numpy.linspace(0, 1, order+1)[numpy.newaxis],
        ...             1./numpy.full(order+1, order+2))
        >>> my_quad(2)
        (array([[0. , 0.5, 1. ]]), array([0.25, 0.25, 0.25]))
        >>> scale_quadrature(my_quad, 2, lower=0, upper=2)
        (array([[0., 1., 2.]]), array([0.5, 0.5, 0.5]))
        >>> scale_quadrature(my_quad, 2, lower=-0.5, upper=0.5)
        (array([[-0.5,  0. ,  0.5]]), array([0.25, 0.25, 0.25]))

    """
    abscissas, weights = quad_func(order=order, **kwargs)
    assert numpy.all(abscissas >= 0) and numpy.all(abscissas <= 1)
    assert numpy.sum(weights) <= 1+1e-10
    assert numpy.sum(weights > 0)
    weights = weights*(upper-lower)
    abscissas = (abscissas.T*(upper-lower)+lower).T
    return abscissas, weights



import numpy
import chaospy


def laguerre(
        order,
        alpha=0.,
        physicist=False,
        normed=False,
        retall=False,
):
    """
    Examples:
        >>> polynomials, norms = chaospy.expansion.laguerre(3, retall=True)
        >>> polynomials
        polynomial([1.0, q0-1.0, q0**2-4.0*q0+2.0, q0**3-9.0*q0**2+18.0*q0-6.0])
        >>> norms
        array([ 1.,  1.,  4., 36.])
        >>> chaospy.expansion.laguerre(3, physicist=True).round(5)
        polynomial([1.0, -q0+1.0, 0.5*q0**2-2.0*q0+2.0,
                    -0.16667*q0**3+1.5*q0**2-5.33333*q0+4.66667])
        >>> chaospy.expansion.laguerre(3, alpha=2, normed=True).round(3)
        polynomial([1.0, 0.577*q0-1.732, 0.204*q0**2-1.633*q0+2.449,
                    0.053*q0**3-0.791*q0**2+3.162*q0-3.162])

    """
    multiplier = -1./numpy.arange(1, order+1) if physicist else 1.
    _, [polynomials], [norms] = chaospy.recurrence.analytical_stieltjes(
        order, chaospy.Gamma(alpha+1), multiplier=multiplier)
    if normed:
        polynomials = chaospy.true_divide(polynomials, numpy.sqrt(norms))
        norms[:] = 1.
    return (polynomials, norms) if retall else polynomials



"""

"""
import numpy
import numpoly


def fit_quadrature(
        orth,
        nodes,
        weights,
        solves,
        retall=False,
        norms=None
):
    """
    Fit polynomial chaos expansion using spectral projection.

    Create a polynomial approximation model from orthogonal expansion,
    quadrature nodes and weights.

    Args:
        orth (numpoly.ndpoly):
            Orthogonal polynomial expansion. Must be orthogonal for the
            approximation to be accurate.
        nodes (numpy.ndarray):
            Where to evaluate the polynomial expansion and model to
            approximate. ``nodes.shape==(D, K)`` where ``D`` is the number of
            dimensions and ``K`` is the number of nodes.
        weights (numpy.ndarray):
            Weights when doing numerical integration. ``weights.shape == (K,)``
            must hold.
        solves (numpy.ndarray):
            The model evaluation to approximate. If `numpy.ndarray` is
            provided, it must have ``len(solves) == K``.
        retall (int):
            What the function should return.
            0: only return fitted polynomials, with shape `evals.shape[1:]`.
            1: polynomials, and Fourier coefficients,
            2: polynomials, coefficients and polynomial evaluations.
        norms (numpy.ndarray):
            Three terms recurrence method produces norms more stable than the
            ones calculated from the polynomials themselves. Calculated from
            quadrature if not provided. ``norms.shape == (len(orth),)`` must
            hold.

    Returns:
        (numpoly.ndpoly):
            Fitted model approximation in the form of an polynomial.
    """
    orth = numpoly.polynomial(orth)
    assert orth.ndim == 1
    weights = numpy.asfarray(weights)
    assert weights.ndim == 1
    solves = numpy.asfarray(solves)
    nodes = numpy.atleast_2d(nodes)
    assert nodes.ndim == 2
    assert nodes.shape[1] == len(weights) == len(solves)

    shape = solves.shape[1:]
    solves = solves.reshape(len(solves), -1)

    ovals = orth(*nodes)
    vals1 = [(val*solves.T*weights).T for val in ovals]

    if norms is None:
        norms = numpy.sum(ovals**2*weights, -1)
    norms = numpy.asfarray(norms)
    assert norms.ndim == 1

    coeffs = (numpy.sum(vals1, 1).T/norms).T
    coeffs = coeffs.reshape(len(coeffs), *shape)
    approx_model = numpoly.sum(orth*coeffs.T, -1).T

    choices = {0: approx_model,
               1: (approx_model, coeffs),
               2: (approx_model, coeffs, ovals)}
    return choices[retall]



"""Hermite orthogonal polynomial expansion."""
import numpy
import chaospy


def hermite(
        order,
        mu=0.,
        sigma=1.,
        physicist=False,
        normed=False,
        retall=False,
):
    """
    Hermite orthogonal polynomial expansion.

    Args:
        order (int):
            The quadrature order.
        mu (float):
            Non-centrality parameter.
        sigma (float):
            Scale parameter.
        physicist (bool):
            Use physicist weights instead of probabilist variant.
        normed (bool):
            If True orthonormal polynomials will be used.
        retall (bool):
            If true return numerical stabilized norms as well. Roughly the same
            as ``cp.E(orth**2, dist)``.

    Returns:
        (numpoly.ndpoly, numpy.ndarray):
            Hermite polynomial expansion. Norms of the orthogonal
            expansion on the form ``E(orth**2, dist)``.

    Examples:
        >>> polynomials, norms = chaospy.expansion.hermite(4, retall=True)
        >>> polynomials
        polynomial([1.0, q0, q0**2-1.0, q0**3-3.0*q0, q0**4-6.0*q0**2+3.0])
        >>> norms
        array([ 1.,  1.,  2.,  6., 24.])
        >>> chaospy.expansion.hermite(3, physicist=True)
        polynomial([1.0, 2.0*q0, 4.0*q0**2-2.0, 8.0*q0**3-12.0*q0])
        >>> chaospy.expansion.hermite(3, sigma=2.5, normed=True).round(3)
        polynomial([1.0, 0.4*q0, 0.113*q0**2-0.707, 0.026*q0**3-0.49*q0])

    """
    multiplier = 2 if physicist else 1
    _, [polynomials], [norms] = chaospy.recurrence.analytical_stieltjes(
        order, chaospy.Normal(mu, sigma), multiplier=multiplier)
    if normed:
        polynomials = chaospy.true_divide(polynomials, numpy.sqrt(norms))
        norms[:] = 1.
    return (polynomials, norms) if retall else polynomials



"""Create samples from the Hammersley set."""
import numpy

from .halton import create_halton_samples


def create_hammersley_samples(order, dim=1, burnin=-1, primes=()):
    """
    Create samples from the Hammersley set.

    The Hammersley set is equivalent to the Halton sequence, except for one
    dimension is replaced with a regular grid.

    Args:
        order (int):
            The order of the Hammersley sequence. Defines the number of samples.
        dim (int):
            The number of dimensions in the Hammersley sequence.
        burnin (int):
            Skip the first ``burnin`` samples. If negative, the maximum of
            ``primes`` is used.
        primes (tuple):
            The (non-)prime base to calculate values along each axis. If
            empty, growing prime values starting from 2 will be used.

    Returns:
        (numpy.ndarray):
            Hammersley set with ``shape == (dim, order)``.

    Examples:
        >>> distribution = chaospy.J(chaospy.Uniform(0, 1), chaospy.Uniform(0, 1))
        >>> samples = distribution.sample(3, rule="hammersley")
        >>> samples.round(4)
        array([[0.75 , 0.125, 0.625],
               [0.25 , 0.5  , 0.75 ]])
        >>> samples = distribution.sample(4, rule="hammersley")
        >>> samples.round(4)
        array([[0.75 , 0.125, 0.625, 0.375],
               [0.2  , 0.4  , 0.6  , 0.8  ]])

    """
    out = numpy.empty((dim, order), dtype=float)
    out[:max(dim-1, 1)] = create_halton_samples(
        order=order, dim=max(dim-1, 1), burnin=burnin, primes=primes)
    if dim > 1:
        out[dim-1] = numpy.linspace(0, 1, order+2)[1:-1]
    return out



import numpy
import chaospy


def gegenbauer(
        order,
        alpha,
        lower=-1,
        upper=1,
        physicist=False,
        normed=False,
        retall=False,
):
    """
    Gegenbauer polynomials.

    Args:
        order (int):
            The polynomial order.
        alpha (float):
            Gegenbauer shape parameter.
        lower (float):
            Lower bound for the integration interval.
        upper (float):
            Upper bound for the integration interval.
        physicist (bool):
            Use physicist weights instead of probabilist.

    Examples:
        >>> polynomials, norms = chaospy.expansion.gegenbauer(4, 1, retall=True)
        >>> polynomials
        polynomial([1.0, q0, q0**2-0.25, q0**3-0.5*q0, q0**4-0.75*q0**2+0.0625])
        >>> norms
        array([1.        , 0.25      , 0.0625    , 0.015625  , 0.00390625])
        >>> chaospy.expansion.gegenbauer(3, 1, physicist=True)
        polynomial([1.0, 2.0*q0, 4.0*q0**2-0.5, 8.0*q0**3-2.0*q0])
        >>> chaospy.expansion.gegenbauer(3, 1, lower=0.5, upper=1.5, normed=True).round(3)
        polynomial([1.0, 4.0*q0-4.0, 16.0*q0**2-32.0*q0+15.0,
                    64.0*q0**3-192.0*q0**2+184.0*q0-56.0])

    """
    multiplier = 1
    if physicist:
        multiplier = numpy.arange(1, order+1)
        multiplier = 2*(multiplier+alpha-1)/multiplier
    _, [polynomials], [norms] = chaospy.recurrence.analytical_stieltjes(
        order, chaospy.Beta(alpha+0.5, alpha+0.5, lower, upper), multiplier=multiplier)
    if normed:
        polynomials = chaospy.true_divide(polynomials, numpy.sqrt(norms))
        norms[:] = 1.
    return (polynomials, norms) if retall else polynomials



"""Fit a polynomial chaos expansion using linear regression."""
import numpy
import numpoly


def fit_regression(
        polynomials,
        abscissas,
        evals,
        model=None,
        retall=0,
):
    """
    Fit a polynomial chaos expansion using linear regression.

    Args:
        polynomials (numpoly.ndpoly):
            Polynomial expansion with ``polynomials.shape == (M,)`` and
            `polynomials.dim=D`.
        abscissas (numpy.ndarray):
            Collocation nodes with ``abscissas.shape == (D, K)``.
        evals (numpy.ndarray):
            Model evaluations with ``len(evals) == K``.
        model (Optional[sklearn.base.BaseEstimator]):
            By default regression is done using the classical least-square
            method. However, if provided, and `sklearn` regression model can be
            used instead.
        retall (int):
            What the function should return.
            0: only return fitted polynomials, with shape `evals.shape[1:]`.
            1: polynomials, and Fourier coefficients,
            2: polynomials, coefficients and polynomial evaluations.

    Returns:
        (chaospy.ndpoly, numpy.ndarray, numpy.ndarray):
            Returned value as determined by `retval`.

    Examples:
        >>> x, y = chaospy.variable(2)
        >>> polynomials = chaospy.polynomial([1, x, y])
        >>> abscissas = [[-1, -1, 1, 1],  [-1, 1, -1, 1]]
        >>> evals = [0, 1, 1, 2]
        >>> chaospy.fit_regression(polynomials, abscissas, evals).round(14)
        polynomial(0.5*q1+0.5*q0+1.0)
        >>> model = sklearn.linear_model.LinearRegression(fit_intercept=False)
        >>> chaospy.fit_regression(
        ...     polynomials, abscissas, evals, model=model).round(14)
        polynomial(0.5*q1+0.5*q0+1.0)

    """
    abscissas = numpy.atleast_2d(abscissas)
    assert abscissas.ndim == 2, "too many dimensions"

    polynomials = numpoly.aspolynomial(polynomials)

    evals = numpy.asarray(evals)
    assert abscissas.shape[-1] == len(evals)

    poly_evals = polynomials(*abscissas).T
    shape = evals.shape[1:]
    if shape:
        evals = evals.reshape(len(evals), -1)

    if model is None:
        uhat, _, _, _ = numpy.linalg.lstsq(poly_evals, evals, rcond=None)

    else:
        try:
            from sklearn.base import BaseEstimator
        except ImportError:  # pragma: no cover
            raise ValueError(
                "arg model != None requires that scikit-learn is installed")
        assert isinstance(model, BaseEstimator), (
            "model not recognized; "
            "Optional[sklearn.base.BaseEstimator] expected"
        )
        if hasattr(model, "fit_intercept"):
            assert not model.fit_intercept, (
                "requires %s(fit_intercept=False)" % model.__class__.__name__)
        uhat = numpy.transpose(model.fit(poly_evals, evals).coef_)

    approx_model = numpoly.sum((polynomials*uhat.T), -1).reshape(shape)
    choices = {0: approx_model,
               1: (approx_model, uhat),
               2: (approx_model, uhat, poly_evals)}
    return choices[retall]



"""Generate antithetic variables."""
import numpy


def create_antithetic_variates(samples, axes=()):
    """
    Generate antithetic variables.

    Args:
        samples (numpy.ndarray):
            The samples, assumed to be on the [0, 1]^D hyper-cube, to be
            reflected.
        axes (tuple):
            Boolean array of which axes to reflect. If This to limit the number
            of points created in higher dimensions by reflecting all axes at
            once.

    Returns (numpy.ndarray):
        Same as ``samples``, but with samples internally reflected. roughly
        equivalent to ``numpy.vstack([samples, 1-samples])`` in one dimensions.
    """
    samples = numpy.asfarray(samples)
    assert numpy.all(samples <= 1) and numpy.all(samples >= 0), (
        "all samples assumed on interval [0, 1].")
    if len(samples.shape) == 1:
        samples = samples.reshape(1, -1)
    inverse_samples = 1-samples
    dims = len(samples)

    if not len(axes):
        axes = (True,)
    axes = numpy.asarray(axes, dtype=bool).flatten()

    indices = {tuple(axes*idx) for idx in numpy.ndindex((2,)*dims)}
    indices = sorted(indices, reverse=True)
    indices = sorted(indices, key=lambda idx: sum(idx))
    out = [numpy.where(idx, inverse_samples.T, samples.T).T for idx in indices]
    out = numpy.dstack(out).reshape(dims, -1)
    return out



# pylint: disable=wildcard-import
"""Module defining distributions."""
from .baseclass import *
from .sampler import *
from .collection import *
from .copulas import *
from .operators import *
from .approximation import *
from .kernel import *

from . import (
    baseclass, sampler, approximation,
    copulas, collection, operators
)



"""Distribution for the core probability distribution."""
import itertools
import numpy
import chaospy

from .distribution import Distribution


class SimpleDistribution(Distribution):
    """Distribution for the core probability distribution."""

    def __init__(
            self,
            parameters=None,
            rotation=None,
            exclusion=None,
            repr_args=None,
    ):
        """
        Args:
            parameters (Optional[Distribution[str, Union[ndarray, Distribution]]]):
                Collection of model parameters.
            rotation (None, Sequence[int], Sequence[Sequence[bool]]):
                The order of which to resolve conditionals. Either as
                a sequence of column rotations, or as a permutation
                matrix. Defaults to `range(len(distribution))` which
                is the same as `p(x0), p(x1|x0), p(x2|x0,x1), ...`.
            repr_args (Optional[Sequence[str]]):
                Positional arguments to place in the object string
                representation. The repr output will then be:
                `<class name>(<arg1>, <arg2>, ...)`.

        """
        if parameters is None:
            parameters = {}
        dependencies, parameters, rotation = chaospy.declare_dependencies(
            distribution=self,
            parameters=parameters,
            rotation=rotation,
        )
        super(SimpleDistribution, self).__init__(
            parameters=parameters,
            dependencies=dependencies,
            rotation=rotation,
            exclusion=exclusion,
            repr_args=repr_args,
        )


    def get_parameters(self, idx, cache, assert_numerical=True):
        """
        Get distribution parameters.

        Uses the cache to replace parameters that are distribution with cached
        values.

        Args:
            cache (Dict[Distribution, numpy.ndarray]):
                Collection of already calculated values.

        Returns:
            (Dict[str, numpy.ndarray]):
                Collection of parameters. Probability distributions are
                replaced with cached values.

        Raise:
            UnsupportedFeature:
                If a parameter is a probability distribution without cache, it
                means the dependency is unresolved.

        """
        parameters = super(SimpleDistribution, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        del parameters["cache"]
        del parameters["idx"]
        for key, value in parameters.items():
            if isinstance(value, Distribution):
                value = value._get_cache(idx, cache, get=0)
                if isinstance(value, Distribution) and assert_numerical:
                    raise chaospy.UnsupportedFeature(
                        "operation not supported for %s with dependencies" % self)
                parameters[key] = value
        return parameters

    def get_upper_parameters(self, idx, cache):
        parameters = self.get_parameters(idx=idx, cache=cache, assert_numerical=False)
        for key, value in parameters.items():
            if isinstance(value, Distribution):
                parameters[key] = value._get_upper(idx, cache)
        return parameters

    def get_lower_parameters(self, idx, cache):
        parameters = self.get_parameters(idx=idx, cache=cache, assert_numerical=False)
        for key, value in parameters.items():
            if isinstance(value, Distribution):
                parameters[key] = value._get_lower(idx, cache)
        return parameters

    def get_mom_parameters(self):
        parameters = self.get_parameters(
            idx=None, cache={}, assert_numerical=False)
        if any([isinstance(value, Distribution) for value in parameters.values()]):
            raise chaospy.UnsupportedFeature(
                "operation not supported for %s with dependencies" % self)
        return parameters

    def _mom(self, kloc, **kwargs):
        """Default moment generator, throws error."""
        raise chaospy.UnsupportedFeature(
            "%s: does not support analytical raw moments." % self)

    def _ttr(self, kloc, **kwargs):
        """Default TTR generator, throws error."""
        raise chaospy.UnsupportedFeature(
            "%s: does not support analytical recurrence coefficients." % self)



# -*- coding: utf-8 -*-
"""Generate the quadrature abscissas and weights in Fejér type I quadrature."""
try:
    from functools import lru_cache
except ImportError:  # pragma: no cover
    from functools32 import lru_cache
import numpy

from .hypercube import hypercube_quadrature


def fejer_1(order, domain=(0, 1), growth=False, segments=1):
    """
    Generate the quadrature abscissas and weights in Fejér type I quadrature.

    Fejér proposed two quadrature rules very similar to
    :func:`chaospy.quadrature.clenshaw_curtis`, except that it does not
    include endpoints, making it more suitable to integrate unbound interval.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.Distribution, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            If True sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.
        segments (int):
            Split intervals into N subintervals and create a patched
            quadrature based on the segmented quadrature. Can not be lower than
            `order`. If 0 is provided, default to square root of `order`.
            Nested samples only exist when the number of segments are fixed.

    Returns:
        abscissas (numpy.ndarray):
            The quadrature points for where to evaluate the model function
            with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
            number of samples.
        weights (numpy.ndarray):
            The quadrature weights with ``weights.shape == (N,)``.

    Notes:
        Implemented as proposed by Waldvogel :cite:`waldvogel_fast_2006`.

    Example:
        >>> abscissas, weights = chaospy.quadrature.fejer_1(3, (0, 1))
        >>> abscissas.round(4)
        array([[0.0381, 0.3087, 0.6913, 0.9619]])
        >>> weights.round(4)
        array([0.1321, 0.3679, 0.3679, 0.1321])

    See also:
        :func:`chaospy.quadrature.gaussian`
        :func:`chaospy.quadrature.clenshaw_curtis`
        :func:`chaospy.quadrature.fejer_2`

    """
    order = numpy.asarray(order)
    order = numpy.where(growth, 2*3**order-1, order)
    return hypercube_quadrature(
        quad_func=fejer_1_simple,
        order=order,
        domain=domain,
        segments=segments,
    )


@lru_cache(None)
def fejer_1_simple(order):
    """Backend for Fejer type I quadrature."""
    order = int(order)
    if order == 0:
        return numpy.array([.5]), numpy.array([1.])
    order += 1

    abscissas = -0.5*numpy.cos(numpy.pi*(numpy.arange(order)+0.5)/order)+0.5

    steps = numpy.arange(1, order, 2)
    length = len(steps)
    remains = order-length

    kappa = numpy.arange(remains)
    beta = numpy.hstack([2*numpy.exp(1j*numpy.pi*kappa/order)/(1-4*kappa**2),
                         numpy.zeros(length+1)])
    beta = beta[:-1]+numpy.conjugate(beta[:0:-1])

    weights = numpy.fft.ifft(beta)
    assert max(weights.imag) < 1e-15
    weights = weights.real/2.

    return abscissas, weights



"""
Generate the quadrature abscissas and weights for simple grid.

Mostly available to ensure that discrete distributions works along side
continuous ones.
"""
import numpy
import chaospy

from .hypercube import hypercube_quadrature


def grid(order, domain=(0, 1), growth=False, segments=1):
    """
    Generate the quadrature abscissas and weights for simple grid.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (chaospy.distributions.baseclass.Distribution, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            The quadrature points and weights. The points are
            equi-spaced grid on the interior of the domain bounds.
            The weights are all equal to `1/len(weights[0])`.

    Example:
        >>> abscissas, weights = chaospy.quadrature.grid(4, chaospy.Uniform(-1, 1))
        >>> abscissas.round(4)
        array([[-0.8, -0.4,  0. ,  0.4,  0.8]])
        >>> weights.round(4)
        array([0.2, 0.2, 0.2, 0.2, 0.2])
        >>> abscissas, weights = chaospy.quadrature.grid([1, 1])
        >>> abscissas.round(4)
        array([[0.25, 0.25, 0.75, 0.75],
               [0.25, 0.75, 0.25, 0.75]])
        >>> weights.round(4)
        array([0.25, 0.25, 0.25, 0.25])

    """
    order = numpy.asarray(order)
    order = numpy.where(growth, numpy.where(order > 0, 3**order-1, 0), order)
    return hypercube_quadrature(
        quad_func=grid_simple,
        order=order,
        domain=domain,
        segments=segments,
    )


def grid_simple(order):
    """
    Backend for grid quadrature.

    Use :func:`chaospy.quadrature.grid` instead.
    """
    order = int(order)
    abscissas = numpy.linspace(0, 1, 2*order+3)[1::2]
    weights = numpy.full(order+1, 1./(order+1))
    return abscissas, weights



"""
Generate the quadrature nodes and weights in Gauss-Radau quadrature.

Example usage
-------------

With increasing order::

    >>> distribution = chaospy.Beta(2, 2, lower=-1, upper=1)
    >>> for order in range(4):  # doctest: +NORMALIZE_WHITESPACE
    ...     abscissas, weights = chaospy.generate_quadrature(
    ...         order, distribution, rule="radau")
    ...     print(abscissas.round(2), weights.round(2))
    [[-1.]] [1.]
    [[-1.   0.2]] [0.17 0.83]
    [[-1.   -0.51  0.13  0.71]] [0.02 0.33 0.48 0.17]
    [[-1.   -0.74 -0.35  0.1   0.53  0.85]]
     [0.01 0.11 0.28 0.34 0.21 0.05]

Multivariate samples::

    >>> distribution = chaospy.J(
    ...     chaospy.Uniform(0, 1), chaospy.Beta(4, 5))
    >>> abscissas, weights = chaospy.generate_quadrature(
    ...     1, distribution, rule="radau")
    >>> abscissas.round(3)
    array([[0.   , 0.   , 0.667, 0.667],
           [0.   , 0.5  , 0.   , 0.5  ]])
    >>> weights.round(3)
    array([0.028, 0.222, 0.083, 0.667])

To change the fixed point, the direct generating function has to be used::

    >>> distribution = chaospy.Uniform(lower=-1, upper=1)
    >>> for fixed_point in numpy.linspace(-1, 1, 6):
    ...     abscissas, weights = chaospy.quadrature.radau(
    ...         2, distribution, fixed_point)
    ...     print(abscissas.round(2), weights.round(2))
    [[-1.   -0.58  0.18  0.82]] [0.06 0.33 0.39 0.22]
    [[-1.04 -0.6   0.17  0.82]] [0.05 0.33 0.39 0.22]
    [[-0.83 -0.2   0.54  0.96]] [0.22 0.38 0.32 0.08]
    [[-0.96 -0.54  0.2   0.83]] [0.08 0.32 0.38 0.22]
    [[-0.82 -0.17  0.6   1.04]] [0.22 0.39 0.33 0.05]
    [[-0.82 -0.18  0.58  1.  ]] [0.22 0.39 0.33 0.06]

However, a fixed point at 0 is not allowed::

    >>> chaospy.quadrature.radau(  # doctest: +IGNORE_EXCEPTION_DETAIL
    ...     3, distribution, fixed_point=0)
    Traceback (most recent call last):
        ...
    numpy.linalg.LinAlgError: Illegal Radau fixed point: 0.0
"""
import numpy
import scipy.linalg
import chaospy

from .utils import combine_quadrature


def radau(
        order,
        dist,
        fixed_point=None,
        recurrence_algorithm="stieltjes",
        rule="clenshaw_curtis",
        tolerance=1e-10,
        scaling=3,
        n_max=5000,
):
    """
    Generate the quadrature nodes and weights in Gauss-Radau quadrature.

    Gauss-Radau formula for numerical estimation of integrals. It requires
    :math:`m+1` points and fits all Polynomials to degree :math:`2m`, so it
    effectively fits exactly all Polynomials of degree :math:`2m-1`.

    It allows for a single abscissas to be user defined, while the others are
    built around this point.

    Canonically, Radau is built around Legendre weight function with the fixed
    point at the left end. Not all distributions/fixed point combinations
    allows for the building of a quadrature scheme.

    Args:
        order (int):
            Quadrature order.
        dist (:class:`chaospy.Distribution`):
            The distribution weights to be used to create higher order nodes
            from.
        fixed_point (float):
            Fixed point abscissas assumed to be included in the quadrature. If
            omitted, use distribution lower bound.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        accuracy (int):
            In the case ``rule`` is used, defines the quadrature order of the
            scheme used. In practice, must be at least as large as ``order``.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.

    Returns:
        abscissas (numpy.ndarray):
            The quadrature points for where to evaluate the model function
            with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
            number of samples.
        weights (numpy.ndarray):
            The quadrature weights with ``weights.shape == (N,)``.

    Example:
        >>> distribution = chaospy.Uniform(-1, 1)
        >>> abscissas, weights = chaospy.quadrature.radau(4, distribution)
        >>> abscissas.round(3)
        array([[-1.   , -0.887, -0.64 , -0.295,  0.094,  0.468,  0.771,  0.955]])
        >>> weights.round(3)
        array([0.016, 0.093, 0.152, 0.188, 0.196, 0.174, 0.125, 0.057])

    """
    if fixed_point is None:
        fixed_point = dist.lower
    else:
        fixed_point = numpy.ones(len(dist))*fixed_point

    if order == 0:
        return fixed_point.reshape(-1, 1), numpy.ones(1)

    coefficients = chaospy.construct_recurrence_coefficients(
        order=2*order-1,
        dist=dist,
        recurrence_algorithm=recurrence_algorithm,
        rule=rule,
        tolerance=tolerance,
        scaling=scaling,
        n_max=n_max,
    )
    coefficients = [radau_jakobi(coeffs, point)
                    for point, coeffs in zip(fixed_point, coefficients)]

    abscissas, weights = chaospy.coefficients_to_quadrature(coefficients)

    return combine_quadrature(abscissas, weights)


def radau_jakobi(coeffs, fixed_point):
    """
    Compute the Radau coefficients.

    Based on the section 7 of the paper "Some modified matrix eigenvalue
    problems", G. Golub.

    Args:
        coeffs_a (numpy.ndarray):
            The first three terms recurrence coefficients of the Gaussian
            quadrature rule.
        coeffs_b (numpy.ndarray):
            The second three terms recurrence coefficients of the Gaussian
            quadrature rule.
        fixed_point (float):
            Fixed point abscissas assumed to be included in the quadrature.

    Returns:
        Three terms recurrence coefficients of the Gauss-Radau quadrature rule.

    Raises:
        numpy.linalg.LinAlgError:
            Error raised if fixed point causes the algorithm to break down.
    """
    right_hand_side = numpy.zeros(len(coeffs[0])-1)
    right_hand_side[-1] = coeffs[1][-1]
    bands_a = numpy.vstack([numpy.sqrt(coeffs[1]), coeffs[0]-fixed_point])
    bands_j = numpy.vstack((bands_a[:, 0:-1], bands_a[0, 1:]))
    try:
        delta = scipy.linalg.solve_banded((1, 1), bands_j, right_hand_side)
    except numpy.linalg.LinAlgError:
        raise numpy.linalg.LinAlgError(
            "Illegal Radau fixed point: %s" % fixed_point.item())
    coeffs = coeffs.copy()
    coeffs[0, -1] = fixed_point+delta[-1]
    return coeffs



"""Create Van Der Corput low discrepancy sequence samples."""
from __future__ import division
import numpy


def create_van_der_corput_samples(idx, number_base=2):
    """
    Create Van Der Corput low discrepancy sequence samples.

    A van der Corput sequence is an example of the simplest one-dimensional
    low-discrepancy sequence over the unit interval; it was first described in
    1935 by the Dutch mathematician J. G. van der Corput. It is constructed by
    reversing the base-n representation of the sequence of natural numbers
    :math:`(1, 2, 3, ...)`.

    In practice, use Halton sequence instead of Van Der Corput, as it is the
    same, but generalized to work in multiple dimensions.

    Args:
        idx (int, numpy.ndarray):
            The index of the sequence. If array is provided, all values in
            array is returned.
        number_base (int):
            The numerical base from where to create the samples from.

    Returns:
        (numpy.ndarray):
            Van der Corput samples.

    Examples:
        #>>> chaospy.create_van_der_corput_samples(range(11), number_base=10)
        #array([0.1 , 0.2 , 0.3 , 0.4 , 0.5 , 0.6 , 0.7 , 0.8 , 0.9 , 0.01, 0.11])
        #>>> chaospy.create_van_der_corput_samples(range(8), number_base=2)
        #array([0.5   , 0.25  , 0.75  , 0.125 , 0.625 , 0.375 , 0.875 , 0.0625])

    """
    assert number_base > 1

    idx = numpy.asarray(idx).flatten() + 1
    out = numpy.zeros(len(idx), dtype=float)

    base = float(number_base)
    active = numpy.ones(len(idx), dtype=bool)
    while numpy.any(active):
        out[active] += (idx[active] % number_base)/base
        idx //= number_base
        base *= number_base
        active = idx > 0
    return out



"""Chebyshev polynomials of the first kind."""
import numpy
import chaospy


def chebyshev_1(
        order,
        lower=-1,
        upper=1,
        physicist=False,
        normed=False,
        retall=False,
):
    """
    Chebyshev polynomials of the first kind.

    Args:
        order (int):
            The polynomial order.
        lower (float):
            Lower bound for the integration interval.
        upper (float):
            Upper bound for the integration interval.
        physicist (bool):
            Use physicist weights instead of probabilist.

    Returns:
        (numpoly.ndpoly, numpy.ndarray):
            Chebyshev polynomial expansion. Norms of the orthogonal
            expansion on the form ``E(orth**2, dist)``.

    Examples:
        >>> polynomials, norms = chaospy.expansion.chebyshev_1(4, retall=True)
        >>> polynomials
        polynomial([1.0, q0, q0**2-0.5, q0**3-0.75*q0, q0**4-q0**2+0.125])
        >>> norms
        array([1.       , 0.5      , 0.125    , 0.03125  , 0.0078125])
        >>> chaospy.expansion.chebyshev_1(3, physicist=True)
        polynomial([1.0, q0, 2.0*q0**2-1.0, 4.0*q0**3-2.5*q0])
        >>> chaospy.expansion.chebyshev_1(3, lower=0.5, upper=1.5, normed=True).round(3)
        polynomial([1.0, 2.828*q0-2.828, 11.314*q0**2-22.627*q0+9.899,
                    45.255*q0**3-135.765*q0**2+127.279*q0-36.77])

    """
    multiplier = 1+numpy.arange(order).astype(bool) if physicist else 1
    _, [polynomials], [norms] = chaospy.recurrence.analytical_stieltjes(
        order, chaospy.Beta(0.5, 0.5, lower, upper), multiplier=multiplier)
    if normed:
        polynomials = chaospy.true_divide(polynomials, numpy.sqrt(norms))
        norms[:] = 1.
    return (polynomials, norms) if retall else polynomials


def chebyshev_2(
        order,
        lower=-1,
        upper=1,
        physicist=False,
        normed=False,
        retall=False,
):
    """
    Chebyshev polynomials of the second kind.

    Args:
        order (int):
            The quadrature order.
        lower (float):
            Lower bound for the integration interval.
        upper (float):
            Upper bound for the integration interval.
        physicist (bool):
            Use physicist weights instead of probabilist.

    Returns:
        (numpoly.ndpoly, numpy.ndarray):
            Chebyshev polynomial expansion. Norms of the orthogonal
            expansion on the form ``E(orth**2, dist)``.

    Examples:
        >>> polynomials, norms = chaospy.expansion.chebyshev_2(4, retall=True)
        >>> polynomials
        polynomial([1.0, q0, q0**2-0.25, q0**3-0.5*q0, q0**4-0.75*q0**2+0.0625])
        >>> norms
        array([1.        , 0.25      , 0.0625    , 0.015625  , 0.00390625])
        >>> chaospy.expansion.chebyshev_2(3, physicist=True)
        polynomial([1.0, 2.0*q0, 4.0*q0**2-0.5, 8.0*q0**3-2.0*q0])
        >>> chaospy.expansion.chebyshev_2(3, lower=0.5, upper=1.5, normed=True).round(3)
        polynomial([1.0, 4.0*q0-4.0, 16.0*q0**2-32.0*q0+15.0,
                    64.0*q0**3-192.0*q0**2+184.0*q0-56.0])

    """
    multiplier = 2 if physicist else 1
    _, [polynomials], [norms] = chaospy.recurrence.analytical_stieltjes(
        order, chaospy.Beta(1.5, 1.5, lower, upper), multiplier=multiplier)
    if normed:
        polynomials= chaospy.true_divide(polynomials, numpy.sqrt(norms))
        norms[:] = 1.
    return (polynomials, norms) if retall else polynomials



"""Constructs distributions for the quantity of interests."""
from __future__ import division
import numpy
import chaospy


def QoI_Dist(poly, dist, sample=10000, **kws):
    """
    Constructs distributions for the quantity of interests.

    The function constructs a kernel density estimator (KDE) for each
    polynomial (poly) by sampling it.  With the KDEs, distributions (Dists) are
    constructed.  The Dists can be used for e.g. plotting probability density
    functions (PDF), or to make a second uncertainty quantification simulation
    with that newly generated Dists.

    Args:
        poly (numpoly.ndpoly):
            Polynomial of interest.
        dist (Distribution):
            Defines the space where the samples for the KDE is taken from the
            poly.
        sample (int):
            Number of samples used in estimation to construct the KDE.

    Returns:
        (Distribution):
            The constructed quantity of interest (QoI) distributions, where
            ``qoi_dists.shape==poly.shape``.

    Examples:
        >>> dist = chaospy.Normal(0, 1)
        >>> x = chaospy.variable(1)
        >>> poly = chaospy.polynomial([x])
        >>> qoi_dist = chaospy.QoI_Dist(poly, dist)
        >>> values = qoi_dist.pdf([-0.75, 0., 0.75])
        >>> values.round(8)
        array([0.29143989, 0.39939823, 0.29531414])

    """
    poly = chaospy.aspolynomial(poly).ravel()
    samples = numpy.atleast_2d(dist.sample(sample, **kws))
    qoi_dist = chaospy.GaussianKDE(poly(*samples))
    return qoi_dist



"""Gauss-Jakobi quadrature rule."""
import numpy
import chaospy

from .hypercube import hypercube_quadrature


def jacobi(order, alpha, beta, lower=-1, upper=1, physicist=False):
    """
    Gauss-Jacobi quadrature rule.

    Compute the sample points and weights for Gauss-Jacobi quadrature. The
    sample points are the roots of the nth degree Jacobi polynomial. These
    sample points and weights correctly integrate polynomials of degree
    :math:`2N-1` or less.

    Gaussian quadrature come in two variants: physicist and probabilist. For
    Gauss-Jacobi physicist means a weight function
    :math:`(1-x)^\alpha (1+x)^\beta` and
    weights that sum to :math`2^{\alpha+\beta}`, and probabilist means a weight
    function is :math:`B(\alpha, \beta) x^{\alpha-1}(1-x)^{\beta-1}` (where
    :math:`B` is the beta normalizing constant) which sum to 1.

    Args:
        order (int):
            The quadrature order.
        alpha (float):
            First Jakobi shape parameter.
        beta (float):
            Second Jakobi shape parameter.
        lower (float):
            Lower bound for the integration interval.
        upper (float):
            Upper bound for the integration interval.
        physicist (bool):
            Use physicist weights instead of probabilist.

    Returns:
        abscissas (numpy.ndarray):
            The ``order+1`` quadrature points for where to evaluate the model
            function with.
        weights (numpy.ndarray):
            The quadrature weights associated with each abscissas.

    Examples:
        >>> abscissas, weights = chaospy.quadrature.jacobi(3, alpha=2, beta=2)
        >>> abscissas
        array([[-0.69474659, -0.25056281,  0.25056281,  0.69474659]])
        >>> weights
        array([0.09535261, 0.40464739, 0.40464739, 0.09535261])

    See also:
        :func:`chaospy.quadrature.gaussian`

    """
    order = int(order)
    coefficients = chaospy.construct_recurrence_coefficients(
        order=order, dist=chaospy.Beta(alpha+1, beta+1, lower, upper))
    [abscissas], [weights] = chaospy.coefficients_to_quadrature(coefficients)
    weights *= 2**(alpha+beta) if physicist else 1
    return abscissas[numpy.newaxis], weights



"""Gauss-Gegenbauer quadrature rule."""
import numpy
import chaospy

from .hypercube import hypercube_quadrature


def gegenbauer(order, alpha, lower=-1, upper=1, physicist=False):
    """
    Gauss-Gegenbauer quadrature rule.

    Compute the sample points and weights for Gauss-Gegenbauer quadrature. The
    sample points are the roots of the nth degree Gegenbauer polynomial. These
    sample points and weights correctly integrate polynomials of degree
    :math:`2N-1` or less.

    Gaussian quadrature come in two variants: physicist and probabilist. For
    Gauss-Gegenbauer physicist means a weight function
    :math:`(1-x^2)^{\alpha-0.5}` and weights that sum to :math`2^{2\alpha-1}`,
    and probabilist means a weight function is
    :math:`B(\alpha+0.5, \alpha+0.5) (x-x^2)^{\alpha+1/2}` (where :math:`B` is
    the beta normalizing constant) which sum to 1.

    Args:
        order (int):
            The quadrature order.
        alpha (float):
            Gegenbauer shape parameter.
        lower (float):
            Lower bound for the integration interval.
        upper (float):
            Upper bound for the integration interval.
        physicist (bool):
            Use physicist weights instead of probabilist.

    Returns:
        abscissas (numpy.ndarray):
            The ``order+1`` quadrature points for where to evaluate the model
            function with.
        weights (numpy.ndarray):
            The quadrature weights associated with each abscissas.

    Examples:
        >>> abscissas, weights = chaospy.quadrature.gegenbauer(3, alpha=2)
        >>> abscissas
        array([[-0.72741239, -0.26621648,  0.26621648,  0.72741239]])
        >>> weights
        array([0.10452141, 0.39547859, 0.39547859, 0.10452141])

    See also:
        :func:`chaospy.quadrature.gaussian`

    """
    order = int(order)
    coefficients = chaospy.construct_recurrence_coefficients(
        order=order, dist=chaospy.Beta(alpha+0.5, alpha+0.5, lower, upper))
    [abscissas], [weights] = chaospy.coefficients_to_quadrature(coefficients)
    weights *= 2**(2*alpha-1) if physicist else 1
    return abscissas[numpy.newaxis], weights



"""Collection of descriptive statistics."""
from .expected import E
from .conditional import E_cond
from .covariance import Cov
from .kurtosis import Kurt
from .percentile import Perc
from .quantity_of_interest import QoI_Dist
from .skewness import Skew
from .standard_deviation import Std
from .variance import Var

from .correlation import *
from .sensitivity import *



"""Discretized Lanczos' method."""
import logging
import numpy
import chaospy


def lanczos(
        order,
        dist,
        rule="clenshaw_curtis",
        tolerance=1e-12,
        scaling=3,
        n_max=1e4,
):
    """
    Discretized Lanczos' method.

    Iterative increase the quadrature order until the norms converges.

    Args:
        order (int):
            The order create recurrence coefficients for.
        dist (chaospy.Distribution):
            The distribution to create recurrence coefficients with respect to.
        rule (str):
            The rule to use to create quadrature nodes and weights from.
        tolerance (float):
            The allowed relative error in norm between two quadrature orders
            before method assumes convergence.
        scaling (float):
            A multiplier the adaptive order increases with for each step
            quadrature order is not converged. Use 0 to indicate unit
            increments.
        n_max (int):
            The allowed number of quadrature points to use in approximation.

    Returns:
        (numpy.ndarray, numpy.ndarray, numpy.ndarray):
            coefficients:
                The recurrence coefficients created using the discretized
                Stieltjes' method, with ``shape == (2, D, order+1)``.

    Examples:
        >>> dist = chaospy.J(chaospy.Beta(3, 6), chaospy.Normal())
        >>> alpha, beta = chaospy.lanczos(3, dist)
        >>> alpha.round(5)
        array([[ 0.33333,  0.39394,  0.42657,  0.44615],
               [-0.     , -0.     , -0.     , -0.     ]])
        >>> beta.round(5)
        array([[1.     , 0.02222, 0.03471, 0.04227],
               [1.     , 1.     , 2.     , 3.     ]])

    Notes:
        The script is adapted from the routine RKPW in W.B. Gragg and W.J.
        Harrod, "The numerically stable reconstruction of Jacobi matrices from
        spectral data", Numer. Math. 44 (1984), 317-335.

    """
    logger = logging.getLogger(__name__)
    if len(dist) > 1:
        assert not dist.stochastic_dependent
        coeffs = zip(*[lanczos(
                order,
                dist_,
                rule=rule,
                tolerance=tolerance,
                scaling=scaling)
        for dist_ in dist])
        coeffs = numpy.vstack(list(coeffs)).reshape((2, len(dist), order+1))
        return coeffs

    order_ = (2*order-1.)/scaling
    beta = beta_old = numpy.nan
    coeffs = numpy.ones((2, order+1))

    while not numpy.all(numpy.abs(beta-beta_old) < tolerance):

        order_ = max(order_*scaling, order_+1)
        if order_ > n_max:
            logger.warning("number of nodes exceeded; stopping with errors:\n%s",
                           ", ".join([numpy.format_float_scientific(val, 1)
                                      for val in numpy.abs(beta-beta_old)/beta]))
            break

        [abscissas], weights = chaospy.generate_quadrature(
            int(order_), dist, rule=rule, segments=0)
        alpha = abscissas[:order+1].copy()
        beta, beta_old = numpy.eye(order+1)[0]*weights[0], beta

        for idx in range(1, len(weights)):

            gamma = 1
            increment_new = 0

            for idy in range(min(idx, order+1)):

                increment = increment_new
                beta_new = gamma*(beta[idy]+weights[idx])
                sigma = 1-gamma
                gamma = numpy.where(
                    beta[idy] <= -weights[idx], 1,
                    beta[idy]/(beta[idy]+weights[idx]))
                increment_new = (
                    (1-gamma)*(alpha[idy]-abscissas[idx])-gamma*increment)
                weights[idx] = numpy.where(
                    gamma >= 1, sigma*beta[idy], increment_new**2/(1-gamma))

                alpha[idy] -= increment_new-increment
                beta[idy] = beta_new

    return numpy.asfarray([alpha.ravel(), beta.ravel()])



"""Baseclass for all Copulas."""
import numpy
import chaospy

from .distribution import Distribution


class CopulaDistribution(Distribution):
    """Baseclass for all Copulas."""

    def __init__(self, dist, trans, rotation=None, repr_args=None):
        r"""
        Args:
            dist (Distribution):
                Distribution to wrap the copula around.
            trans (Distribution):
                The copula wrapper `[0,1]^D \into [0,1]^D`.

        """
        assert len(dist) == len(trans), "Copula length missmatch"
        accumulant = set()
        dependencies = [deps.copy() for deps in dist._dependencies]
        for idx, _ in sorted(enumerate(trans._dependencies), key=lambda x: len(x[1])):
            accumulant.update(dist._dependencies[idx])
            dependencies[idx] = accumulant.copy()

        super(CopulaDistribution, self).__init__(
            parameters=dict(dist=dist, trans=trans),
            dependencies=dependencies,
            rotation=rotation,
            repr_args=repr_args,
        )

    def get_parameters(self, idx, cache, assert_numerical=True):
        parameters = super(CopulaDistribution, self).get_parameters(
            idx, cache, assert_numerical=assert_numerical)
        if idx is None:
            del parameters["idx"]
        return parameters

    def _lower(self, idx, dist, trans, cache):
        return dist._get_lower(idx, cache=cache)

    def _upper(self, idx, dist, trans, cache):
        return dist._get_upper(idx, cache=cache)

    def _cdf(self, xloc, idx, dist, trans, cache):
        output = dist._get_fwd(xloc, idx, cache=cache)
        output = trans._get_fwd(output, idx, cache=cache)
        return output

    def _ppf(self, qloc, idx, dist, trans, cache):
        qloc = trans._get_inv(qloc, idx, cache=cache)
        xloc = dist._get_inv(qloc, idx, cache=cache)
        return xloc

    def _pdf(self, xloc, idx, dist, trans, cache):
        density = dist._get_pdf(xloc, idx, cache=cache.copy())
        return trans._get_pdf(
            dist._get_fwd(xloc, idx, cache=cache), idx, cache=cache)*density



"""Generate quadrature abscissas and weights for discrete distributions."""
import numpy
import chaospy

from .hypercube import hypercube_quadrature


def discrete(order, domain=(0, 1), growth=False):
    """
    Generate quadrature abscissas and weights for discrete distributions.

    A specialized quadrature designed for discrete distributions. It is defined
    as an evenly spaced grid on the domain, rounded to nearest integer. Rule
    will converge to where all integer values on the domain is covered. This
    ensure that only necessary samples are evaluated.

    Args:
        order (int, numpy.ndarray):
            Quadrature order.
        domain (:class:`chaospy.Distribution`, numpy.ndarray):
            Either distribution or bounding of interval to integrate over.
        growth (bool):
            if true sets the growth rule for the quadrature rule to only
            include orders that enhances nested samples.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            The quadrature points and weights. The points are
            equi-spaced grid on the interior of the domain bounds.
            The weights are all equal to `1/len(weights[0])`.
            Either distribution or bounding of interval to integrate over.

    Examples:
        >>> distribution = chaospy.Binomial(6, 0.4)
        >>> abscissas, weights = chaospy.quadrature.discrete(4, distribution)
        >>> abscissas
        array([[0., 2., 3., 4., 6.]])
        >>> weights.round(4)
        array([0.0601, 0.4006, 0.3561, 0.178 , 0.0053])
        >>> abscissas, weights = chaospy.quadrature.discrete(10, distribution)
        >>> abscissas
        array([[0., 1., 2., 3., 4., 5., 6.]])
        >>> weights.round(4)
        array([0.0467, 0.1866, 0.311 , 0.2765, 0.1382, 0.0369, 0.0041])

    """
    order = numpy.asarray(order)
    order = numpy.where(growth, numpy.where(order > 0, 2**order, 0), order)
    return hypercube_quadrature(
        quad_func=discrete_simple,
        order=order,
        domain=domain,
        auto_scale=False,
    )


def discrete_simple(order, lower=-2, upper=2):
    """
    Backend for discrete quadrature.

    Use :func:`chaospy.quadrature.discrete` instead.
    """
    order = int(min(order, round(upper-lower)-1))
    abscissas = numpy.linspace(lower, upper, 2*order+3)[1::2].round()
    weights = numpy.full(order+1, (upper-lower)/(order+1.))
    return abscissas, weights



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chaospy',
 'chaospy.descriptives',
 'chaospy.descriptives.correlation',
 'chaospy.descriptives.sensitivity',
 'chaospy.distributions',
 'chaospy.distributions.baseclass',
 'chaospy.distributions.collection',
 'chaospy.distributions.copulas',
 'chaospy.distributions.kernel',
 'chaospy.distributions.operators',
 'chaospy.distributions.sampler',
 'chaospy.distributions.sampler.sequences',
 'chaospy.expansion',
 'chaospy.external',
 'chaospy.quadrature',
 'chaospy.recurrence']

package_data = \
{'': ['*']}

install_requires = \
['numpoly>=1.1.2,<2.0.0', 'numpy', 'scipy']

setup_kwargs = {
    'name': 'chaospy',
    'version': '4.3.3',
    'description': 'Numerical tool for perfroming uncertainty quantification',
    'long_description': '.. image:: https://github.com/jonathf/chaospy/raw/master/docs/_static/chaospy_logo.svg\n   :height: 200 px\n   :width: 200 px\n   :align: center\n\n|circleci| |codecov| |readthedocs| |downloads| |pypi|\n\n.. |circleci| image:: https://img.shields.io/circleci/build/github/jonathf/chaospy/master\n    :target: https://circleci.com/gh/jonathf/chaospy/tree/master\n.. |codecov| image:: https://img.shields.io/codecov/c/github/jonathf/chaospy\n    :target: https://codecov.io/gh/jonathf/chaospy\n.. |readthedocs| image:: https://img.shields.io/readthedocs/chaospy\n    :target: https://chaospy.readthedocs.io/en/master/?badge=master\n.. |downloads| image:: https://img.shields.io/pypi/dm/chaospy\n    :target: https://pypistats.org/packages/chaospy\n.. |pypi| image:: https://img.shields.io/pypi/v/chaospy\n    :target: https://pypi.org/project/chaospy\n\n* `Documentation <https://chaospy.readthedocs.io/en/master>`_\n* `Interactive tutorials with Binder <https://mybinder.org/v2/gh/jonathf/chaospy/master?filepath=docs%2Fuser_guide>`_\n* `Code of conduct <https://github.com/jonathf/chaospy/blob/master/CODE_OF_CONDUCT.md>`_\n* `Contribution guideline <https://github.com/jonathf/chaospy/blob/master/CONTRIBUTING.md>`_\n* `Changelog <https://github.com/jonathf/chaospy/blob/master/CHANGELOG.md>`_\n* `License <https://github.com/jonathf/chaospy/blob/master/LICENCE.txt>`_\n\nChaospy is a numerical toolbox for performing uncertainty quantification using\npolynomial chaos expansions, advanced Monte Carlo methods implemented in\nPython. It also include a full suite of tools for doing low-discrepancy\nsampling, quadrature creation, polynomial manipulations, and a lot more.\n\nThe philosophy behind ``chaospy`` is not to be a single tool that solves every\nuncertainty quantification problem, but instead be a specific tools to aid to\nlet the user solve problems themselves. This includes both well established\nproblems, but also to be a foundry for experimenting with new problems, that\nare not so well established. To do this, emphasis is put on the following:\n\n* Focus on an easy to use interface that embraces the `pythonic code style\n  <https://docs.python-guide.org/writing/style/>`_.\n* Make sure the code is "composable", such a way that changing one part of the\n  code with something user defined should be easy and encouraged.\n* Try to support a broad width of the various methods for doing uncertainty\n  quantification where that makes sense to involve ``chaospy``.\n* Make sure that ``chaospy`` plays nice with a large set of of other other\n  similar projects. This includes `numpy <https://numpy.org/>`_, `scipy\n  <https://scipy.org/>`_, `scikit-learn <https://scikit-learn.org>`_,\n  `statsmodels <https://statsmodels.org/>`_, `openturns\n  <https://openturns.org/>`_, and `gstools <https://geostat-framework.org/>`_\n  to mention a few.\n* Contribute all code to the community open source.\n\nInstallation\n============\n\nInstallation should be straight forward from `pip <https://pypi.org/>`_:\n\n.. code-block:: bash\n\n    pip install chaospy\n\nOr if `Conda <https://conda.io/>`_ is more to your liking:\n\n.. code-block:: bash\n\n    conda install -c conda-forge chaospy\n\nThen go over to the `documentation <https://chaospy.readthedocs.io/en/master>`_\nto see how to use the toolbox.\n\nDevelopment\n===========\n\nChaospy uses `poetry`_ to manage its development installation. Assuming\n`poetry`_ installed on your system, installing ``chaospy`` for development can\nbe done from the repository root with the command:\n\n.. code-block:: bash\n\n    poetry install\n\nThis will install all required dependencies and chaospy into a virtual\nenvironment.\n\n.. _poetry: https://poetry.eustace.io/\n\nTesting\n-------\n\nTo ensure that the code run on your local system, run the following:\n\n.. code-block:: bash\n\n    poetry run pytest --doctest-modules chaospy/ tests/\n\nDocumentation\n-------------\n\nThe documentation build assumes that ``pandoc`` is installed on your\nsystem and available in your path.\n\nTo build documentation locally on your system, use ``make`` from the ``docs/``\nfolder:\n\n.. code-block:: bash\n\n    cd docs/\n    make html\n\nRun ``make`` without argument to get a list of build targets.\nThe HTML target stores output to the folder ``doc/.build/html``.\n',
    'author': 'Jonathan Feinberg',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jonathf/chaospy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)
